#!/usr/bin/env python3
# thoth-report-processing
# Copyright(C) 2020 Francesco Murdaca
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Security Indicator cloc test suite."""

from .base_test import ReportProcessingTestCase
from thoth.report_processing.components.security import SecurityIndicatorsCloc


class TestSecurityReportsCloc(ReportProcessingTestCase):
    """Test implementation of security indicator cloc."""

    _SI_REPORT_NAME = "security-indicator-54c6daf9"

    _SI_CLOC_FOLDER_PATH = ReportProcessingTestCase.DATA / "security-indicator"

    def test_get_security_indicator_cloc_report(self) -> None:
        """Test retrieving report from local path."""
        si_cloc_reports = SecurityIndicatorsCloc.aggregate_security_indicator_cloc_results(
            security_indicator_cloc_repo_path=self._SI_CLOC_FOLDER_PATH,
            is_local=True,
        )
        assert si_cloc_reports[0]

    def test_get_metadata_df_from_cloc_report(self) -> None:
        """Test obtaining metadata from si cloc report."""
        si_cloc_reports = SecurityIndicatorsCloc.aggregate_security_indicator_cloc_results(
            security_indicator_cloc_repo_path=self._SI_CLOC_FOLDER_PATH,
            is_local=True,
        )
        si_cloc_report = si_cloc_reports[0]
        metadata_retrieved = SecurityIndicatorsCloc()._extract_data_from_si_cloc_metadata(si_cloc_report=si_cloc_report)
        metadata_retrieved_keys = [k for k in metadata_retrieved]
        metadata_test_keys = [
            "datetime_si_cloc",
            "analyzer_si_cloc",
            "analyzer_version_si_cloc",
            "document_id_si_cloc",
            "package_name",
            "package_version",
            "package_index",
        ]
        assert metadata_retrieved_keys == metadata_test_keys



#!/usr/bin/env python3
# thoth-report-processing
# Copyright(C) 2020 Francesco Murdaca
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Amun Inspection reports processing methods."""

import os
import logging
import json
import copy
import hashlib

from pathlib import Path
from typing import List, Optional, Tuple, Dict, Any, Iterable

from sklearn.preprocessing import LabelEncoder

from numpy import array
import numpy as np
import pandas as pd

from thoth.report_processing.exceptions import ThothNotKnownResultStore
from thoth.report_processing.exceptions import ThothMissingDatasetAtPath
from thoth.report_processing.enums import ThothAmunInspectionFileStoreEnum

from thoth.storages.inspections import InspectionStore

# set up logging
DEBUG_LEVEL = bool(int(os.getenv("DEBUG_LEVEL", 0)))

if DEBUG_LEVEL:
    logging.basicConfig(level=logging.DEBUG)
else:
    logging.basicConfig(level=logging.INFO)

_LOGGER = logging.getLogger(__name__)


class AmunInspections:
    """Class of methods used to process reports from Amun Inspections."""

    _INSPECTION_PERFORMANCE_VALUES = {"elapsed_time": "stdout__@result__elapsed", "rate": "stdout__@result__rate"}

    _INSPECTION_USAGE_VALUES = [
        "usage__ru_inblock",
        "usage__ru_majflt",
        "usage__ru_maxrss",
        "usage__ru_minflt",
        "usage__ru_nivcsw",
        "usage__ru_nvcsw",
        "usage__ru_stime",
        "usage__ru_utime",
    ]

    @classmethod
    def aggregate_thoth_inspections_results(
        cls,
        store_files: Optional[List[str]] = None,
        inspections_identifiers: Optional[List[str]] = None,
        inspection_ids_list: Optional[List[str]] = None,
        limit_results: bool = False,
        max_ids: int = 5,
        is_local: bool = False,
        repo_path: Optional[Path] = None,
        store_locally: bool = False,
        store_locally_repo_name: str = "./inspections",
    ) -> Dict[str, Any]:
        """Aggregate results stored on Ceph or locally from repo for Thoth components reports.

        :param store_files: files to be retrieved from the Store for each result, if None all files are retrieved.
        :param inspections_identifiers: Inspection identifiers in inspection IDs.
        :param inspection_ids_list: Inspection IDs list.
        :param limit_results: reduce the number of reports ids considered to `max_ids`.
        :param max_ids: maximum number of reports ids considered.
        :param is_local: flag to retrieve the dataset locally (if not uses Ceph S3 (credentials are required)).
        :param repo_path: required if you want to retrieve the dataset locally and `is_local` is set to True.
        :param store_locally: required if you want to store the dataset retrieved from Ceph.
        :param store_locally_repo_name: repo name where to store the dataset retrieved from Ceph,
                './inspections' by default.
        """
        if store_files:
            if any(store_file not in ThothAmunInspectionFileStoreEnum.__members__ for store_file in store_files):
                raise ThothNotKnownResultStore(
                    f"InspectionStore does not contain some of the files listed: {store_files}."
                    f"InspectionStore: {ThothAmunInspectionFileStoreEnum.__members__.keys()}",
                )

        if limit_results:
            _LOGGER.debug(f"Limiting results to {max_ids}!")

        files: Dict[str, Any] = {}

        if not store_files:
            store_files = ["results", "specification", "hardware_info", "build_logs"]

        if is_local:
            files, counter = cls._aggregate_inspections_from_local(
                repo_path=repo_path,
                inspections_identifiers=inspections_identifiers,
                files=files,
                limit_results=limit_results,
                max_ids=max_ids,
                store_files=store_files,
            )

        else:
            files, counter = cls._aggregate_inspections_from_ceph(
                store_files=store_files,
                inspections_identifiers=inspections_identifiers,
                inspection_ids_list=inspection_ids_list,
                files=files,
                limit_results=limit_results,
                max_ids=max_ids,
                store_locally=store_locally,
                store_locally_repo_name=store_locally_repo_name,
            )

        _LOGGER.info("Number of files retrieved is: %r" % counter)

        return files

    @classmethod
    def _aggregate_inspections_from_local(
        cls,
        files: Dict[str, Any],
        store_files: Optional[List[str]] = None,
        inspections_identifiers: Optional[List[str]] = None,
        repo_path: Optional[Path] = None,
        limit_results: bool = False,
        max_ids: int = 5,
    ) -> Tuple[Dict[str, Any], int]:
        """Aggregate Thoth results from local repo."""
        _LOGGER.info(f"Retrieving dataset at path... {repo_path}")
        if not repo_path:
            _LOGGER.warning("No Path has been provided to retrieve data locally.")
            return files, 0

        if not repo_path.exists():
            raise ThothMissingDatasetAtPath(f"There is no dataset at this path: {repo_path}.")

        counter = 0

        # Iterate through inspection IDs
        for result_path in repo_path.iterdir():
            inspection_document_id = result_path.name

            identifier_check = False

            if inspections_identifiers:

                identifier_check = cls._has_inspection_identifier(
                    inspection_document_id,
                    inspections_identifiers,
                )

            if inspections_identifiers and not identifier_check:
                _LOGGER.info(f"Skipping inspection ID... {inspection_document_id}")
                # If identifiers are requested and inspection id does not contain any of them, skip it
                continue

            _LOGGER.info(f"Considering inspection ID... {inspection_document_id}")

            retrieved_files: List[Dict[str, Any]] = []
            try:
                # Iterate through inspection results number
                for inspection_number_path in Path(f"{result_path}/results").iterdir():
                    _LOGGER.info(
                        f"Considering inspection ID {inspection_document_id}." f"Number {inspection_number_path.name}",
                    )

                    file_info: Dict[str, Any] = {}

                    if store_files and ThothAmunInspectionFileStoreEnum.results.name in store_files:

                        with open(f"{inspection_number_path}/result", "r") as result_file:
                            inspection_result_document = json.load(result_file)

                            file_info["result"] = inspection_result_document
                            file_info["result"]["inspection_document_id"] = inspection_document_id

                        if store_files and ThothAmunInspectionFileStoreEnum.hardware_info.name in store_files:

                            with open(f"{inspection_number_path}/hwinfo", "r") as hwinfo_file:
                                inspection_hw_info = json.load(hwinfo_file)

                                file_info["hwinfo"] = inspection_hw_info

                        if store_files and ThothAmunInspectionFileStoreEnum.job_logs.name in store_files:

                            with open(f"{inspection_number_path}/log", "r") as job_log_file:
                                inspection_job_logs = job_log_file.read()

                                file_info["job_logs"] = inspection_job_logs

                    if file_info:
                        retrieved_files.append(file_info)

            except Exception as retrieval_error:
                _LOGGER.info(
                    f"Considering inspection ID {inspection_document_id}."
                    f"No files retrieved due to the following error: {retrieval_error}",
                )

            if retrieved_files:
                files[inspection_document_id] = {"results": retrieved_files}
            else:
                files[inspection_document_id] = {}

            try:
                if store_files and ThothAmunInspectionFileStoreEnum.specification.name in store_files:

                    with open(f"{result_path}/build/specification", "r") as specification_file:
                        inspection_specification_document = json.load(specification_file)

                        if retrieved_files:
                            modified_results = []
                            for result in files[inspection_document_id]["results"]:
                                result["result"]["identifier"] = inspection_specification_document["identifier"]
                                result["result"]["specification_base"] = inspection_specification_document["base"]
                                result["result"]["batch_size"] = inspection_specification_document["batch_size"]
                                result["requirements"] = inspection_specification_document["python"]["requirements"]

                                requirements_locked = cls._parse_requirements_locked(
                                    requirements_locked=inspection_specification_document["python"][
                                        "requirements_locked"
                                    ],
                                )
                                result["result"]["requirements_locked"] = requirements_locked

                                result["result"]["run"] = inspection_specification_document["run"]

                                modified_results.append(result)

                            files[inspection_document_id] = {"results": modified_results}

                        files[inspection_document_id]["specification"] = inspection_specification_document

            except Exception as retrieval_error:
                _LOGGER.info(
                    f"Considering inspection ID {inspection_document_id}."
                    f"No build specification retrieved due to the following error: {retrieval_error}",
                )

            try:
                if store_files and ThothAmunInspectionFileStoreEnum.build_logs.name in store_files:

                    with open(f"{result_path}/build/log", "r") as build_logs_type:
                        inspection_build_logs = build_logs_type.read()

                        files[inspection_document_id]["build_logs"] = inspection_build_logs

            except Exception as retrieval_error:
                _LOGGER.info(
                    f"Considering inspection ID {inspection_document_id}."
                    f"No build log retrieved due to the following error: {retrieval_error}",
                )

            if files[inspection_document_id]:
                counter += 1

                if limit_results:
                    if counter == max_ids:
                        return files, counter
            else:
                files.pop(inspection_document_id)

        return files, counter

    @staticmethod
    def _has_inspection_identifier(
        inspection_id: str,
        inspections_identifiers: List[str],
    ) -> bool:
        """Check if inspection id has identifier."""
        inspection_id_pieces = inspection_id.split("-")

        for identifier in inspections_identifiers:
            identifier_pieces = identifier.split("-")
            if not set(identifier_pieces) - set(inspection_id_pieces):
                # The inspection id has the correct identifier requested
                return True

        # The inspection id does not has the identifier requested
        return False

    @staticmethod
    def _retrieve_inspection_results_from_ceph(
        inspection_document_id: str,
        inspection_store: InspectionStore,
        store_locally_repo_name: str,
        store_files: Optional[List[str]] = None,
        store_locally: bool = False,
    ) -> Tuple[List[Dict[str, Any]], int]:
        """Retrieve inspection results for inspection ID."""
        _LOGGER.info(f"Document id: {inspection_document_id}")
        number_results = inspection_store.results.get_results_count()

        if number_results == 0:
            _LOGGER.warning(f"No inspection results identified for inspection ID: {inspection_document_id}")
            return [], 0

        # Inspection ID result counter
        inspection_result_counter = 0

        retrieved_files: List[Dict[str, Any]] = []

        for inspection_result_number in range(number_results):

            file_info: Dict[str, Any] = {}

            try:

                if store_files and ThothAmunInspectionFileStoreEnum.results.name in store_files:
                    inspection_result_document = inspection_store.results.retrieve_result(
                        item=inspection_result_number,
                    )

                    file_info["result"] = inspection_result_document
                    file_info["result"]["inspection_document_id"] = inspection_document_id

                    if store_locally:
                        result_path = (
                            f"{store_locally_repo_name}/{inspection_document_id}/results/{inspection_result_number}"
                        )
                        if not os.path.exists(result_path):
                            os.makedirs(result_path)

                    with open(
                        f"{store_locally_repo_name}/{inspection_document_id}/results/{inspection_result_number}/result",
                        "w",
                    ) as result_file:
                        result_file.write(json.dumps(inspection_result_document))

                if store_files and ThothAmunInspectionFileStoreEnum.hardware_info.name in store_files:
                    inspection_hw_info = inspection_store.results.retrieve_hwinfo(
                        item=inspection_result_number,
                    )

                    file_info["hwinfo"] = inspection_hw_info

                    if store_locally:
                        with open(
                            f"{store_locally_repo_name}/{inspection_document_id}/results/{inspection_result_number}/hwinfo",
                            "w",
                        ) as hw_file:
                            hw_file.write(json.dumps(inspection_hw_info))

                if store_files and ThothAmunInspectionFileStoreEnum.job_logs.name in store_files:
                    inspection_job_logs = inspection_store.results.retrieve_log(item=inspection_result_number)

                    file_info["job_logs"] = inspection_job_logs

                    if store_locally:
                        with open(
                            f"{store_locally_repo_name}/{inspection_document_id}/results/{inspection_result_number}/log",
                            "w",
                        ) as log_file:
                            log_file.write(inspection_job_logs)

                inspection_result_counter += 1

                _LOGGER.info(
                    "From inspection id %r results number identifier retrieved: %r",
                    inspection_document_id,
                    inspection_result_number,
                )

                if file_info:
                    retrieved_files.append(file_info)

            except Exception as inspection_exception:
                _LOGGER.exception(
                    f"Exception during retrieval of inspection id {inspection_document_id} results"
                    f"n.{inspection_result_number}: {inspection_exception}",
                )
                pass

        return retrieved_files, inspection_result_number

    @staticmethod
    def _retrieve_build_specification_from_ceph(
        inspection_document_id: str,
        inspection_store: InspectionStore,
        store_locally_repo_name: str,
        store_files: Optional[List[str]] = None,
        store_locally: bool = False,
    ) -> Any:
        """Retrieve inspection build specification for inspection ID."""
        if store_files and ThothAmunInspectionFileStoreEnum.specification.name in store_files:

            try:
                inspection_specification_document = inspection_store.retrieve_specification()

            except Exception as inspection_exception:
                _LOGGER.exception(
                    f"Exception during retrieval of inspection build specification for"
                    f"inspection id {inspection_document_id}: {inspection_exception}",
                )
                return None

            if store_locally:
                with open(
                    f"{store_locally_repo_name}/{inspection_document_id}/build/specification",
                    "w",
                ) as specification_file:
                    specification_file.write(json.dumps(inspection_specification_document))

            return inspection_specification_document

        else:
            return None

    @staticmethod
    def _retrieve_build_logs_from_ceph(
        inspection_document_id: str,
        inspection_store: InspectionStore,
        store_locally_repo_name: str,
        store_files: Optional[List[str]] = None,
        store_locally: bool = False,
    ) -> Any:
        """Retrieve inspection build logs for inspection ID."""
        if store_files and ThothAmunInspectionFileStoreEnum.build_logs.name in store_files:

            try:
                inspection_build_logs = inspection_store.build.retrieve_log()

            except Exception as inspection_exception:
                _LOGGER.exception(
                    f"Exception during retrieval of inspection build logs for"
                    f"inspection id {inspection_document_id}: {inspection_exception}",
                )
                return None

            if store_locally:
                with open(f"{store_locally_repo_name}/{inspection_document_id}/build/log", "w") as build_log_file:
                    build_log_file.write(inspection_build_logs)

            return inspection_build_logs

        else:
            return None

    @classmethod
    def _aggregate_inspections_from_ceph(
        cls,
        files: Dict[str, Any],
        store_locally_repo_name: str,
        store_files: Optional[List[str]] = None,
        inspections_identifiers: Optional[List[str]] = None,
        inspection_ids_list: Optional[List[str]] = None,
        limit_results: bool = False,
        max_ids: int = 5,
        store_locally: bool = False,
    ) -> Tuple[Dict[str, Any], int]:
        """Aggregate Thoth results from Ceph."""
        store_class_type = InspectionStore

        # Inspection ID counter
        files_retrieved_counter = 0

        if store_locally:
            if not os.path.exists(store_locally_repo_name):
                os.mkdir(store_locally_repo_name)

        for inspection_document_id in inspection_ids_list or store_class_type.iter_inspections():

            identifier_check = False

            if inspections_identifiers:

                identifier_check = cls._has_inspection_identifier(
                    inspection_document_id,
                    inspections_identifiers,
                )

            if inspections_identifiers and not identifier_check:
                # If identifiers are requested and inspection id does not contain any of them, skip it
                continue

            inspection_store = InspectionStore(inspection_id=inspection_document_id)
            inspection_store.connect()

            retrieved_files, _ = cls._retrieve_inspection_results_from_ceph(
                inspection_document_id=inspection_document_id,
                inspection_store=inspection_store,
                store_files=store_files,
                store_locally=store_locally,
                store_locally_repo_name=store_locally_repo_name,
            )

            if store_locally:
                build_path = f"{store_locally_repo_name}/{inspection_document_id}/build/"
                if not os.path.exists(build_path):
                    os.makedirs(build_path)

            build_specification_file = cls._retrieve_build_specification_from_ceph(
                inspection_document_id=inspection_document_id,
                inspection_store=inspection_store,
                store_files=store_files,
                store_locally=store_locally,
                store_locally_repo_name=store_locally_repo_name,
            )

            if retrieved_files and build_specification_file:
                modified_results = []

                for result in retrieved_files:
                    result["result"]["identifier"] = build_specification_file["identifier"]
                    result["result"]["specification_base"] = build_specification_file["base"]
                    result["result"]["batch_size"] = build_specification_file["batch_size"]
                    result["requirements"] = build_specification_file["python"]["requirements"]

                    requirements_locked = cls._parse_requirements_locked(
                        requirements_locked=build_specification_file["python"]["requirements_locked"],
                    )
                    result["result"]["requirements_locked"] = requirements_locked

                    result["result"]["run"] = build_specification_file["run"]

                    modified_results.append(result)

                files[inspection_document_id] = {"results": modified_results}
            else:
                files[inspection_document_id] = {}

            if build_specification_file:
                files[inspection_document_id]["specification"] = build_specification_file

            build_info_file = cls._retrieve_build_logs_from_ceph(
                inspection_document_id=inspection_document_id,
                inspection_store=inspection_store,
                store_files=store_files,
                store_locally=store_locally,
                store_locally_repo_name=store_locally_repo_name,
            )

            if build_info_file:
                files[inspection_document_id]["build_logs"] = build_info_file

            if files[inspection_document_id]:
                _LOGGER.info(f"Retrieved file n. {files_retrieved_counter}")
                files_retrieved_counter += 1

                if limit_results:
                    if files_retrieved_counter == max_ids:
                        return files, files_retrieved_counter
            else:
                files.pop(inspection_document_id)

        return files, files_retrieved_counter

    @staticmethod
    def _parse_requirements_locked(requirements_locked: Dict[str, Any]) -> Dict[str, Any]:
        """Parse requirements_locked to make sure name, version, index is present."""
        default = requirements_locked["default"]
        for package_name, data in default.items():

            # Use PyPI index as default if index is missing
            if "index" not in data.keys():
                modified_data = data.copy()
                modified_data["index"] = "pypi"
                requirements_locked["default"][package_name].update(modified_data)

        return requirements_locked

    @classmethod
    def process_inspection_runs(
        cls,
        inspection_runs: Dict[str, Any],
        filter_by_batch_size: int = 1,
    ) -> Tuple[Dict[str, pd.DataFrame], Dict[str, pd.DataFrame]]:
        """Process inspection runs into pd.DataFrame for each inspection ID.

        :param inspection_runs: aggregated data provided by `aggregate_thoth_inspections_runs`.
        :param filter_by_batch_size: filter inspection to guarantee all have same batch size.
        """
        processed_inspection_runs: Dict[str, pd.DataFrame] = {}
        failed_inspection_runs: Dict[str, pd.DataFrame] = {}

        if not inspection_runs:
            _LOGGER.exception("Empty iterable provided.")
            return processed_inspection_runs, failed_inspection_runs

        for inspection_id, inspection_run in inspection_runs.items():

            if "results" not in inspection_run.keys():
                _LOGGER.warning(
                    f"Inspection ID {inspection_id} has not results, discarding...",
                )
                continue

            inspection_run_df = cls.process_inspection_run(inspection_run=inspection_run)

            if any(exit_code != 0 for exit_code in inspection_run_df["exit_code"].values):
                _LOGGER.warning(
                    f"Inspection ID {inspection_id} has batch size of: {inspection_run_df.shape[0]}"
                    f" but some of them are failed.",
                )
                failed_inspection_runs[inspection_id] = inspection_run_df

            elif inspection_run_df.shape[0] >= filter_by_batch_size:
                processed_inspection_runs[inspection_id] = inspection_run_df

            else:
                _LOGGER.warning(
                    f"Inspection ID {inspection_id} has batch size of: {inspection_run_df.shape[0]}"
                    f"... discarding due to filter set to: {filter_by_batch_size}",
                )

        return processed_inspection_runs, failed_inspection_runs

    @staticmethod
    def process_inspection_run(inspection_run: Dict[str, Any]) -> pd.DataFrame:
        """Process an inspection run into pd.DataFrame.

        :param inspection_run: single aggregated inspection ID provided by `aggregate_thoth_inspections_runs`.
        """
        results = inspection_run["results"]
        final_df = pd.DataFrame()

        inspection_numbers = []
        inspection_number = 0

        for inspection in results:
            inspection_result_df = pd.json_normalize(inspection["result"], sep="__")

            final_df = pd.concat([final_df, inspection_result_df], axis=0)
            inspection_numbers.append(inspection_number)
            inspection_number += 1

        final_df.reset_index(inplace=True, drop=True)
        final_df["inspection_number"] = inspection_numbers

        return final_df

    @staticmethod
    def evaluate_statistics_on_inspection_df(
        inspection_df: pd.DataFrame,
        column_names: List[str],
        extra_columns: List[str],
    ) -> pd.DataFrame:
        """Evaluate statistics on performance values selected from Dataframe columns."""
        unashable_columns = inspection_df.applymap(lambda x: isinstance(x, dict) or isinstance(x, list)).all()[
            lambda x: x == True  # noqa
        ]
        new_data = {}

        inspection_start = None
        inspection_end = None
        inspection_duration = None

        for c_name in inspection_df.columns.values:

            if c_name in column_names:
                # TODO: Allow user to select another parameter, median used by default
                new_data[c_name] = [inspection_df[c_name].median()]

            elif c_name == "end_datetime":

                if "start_datetime" in inspection_df.columns.values:
                    inspection_start = pd.to_datetime(inspection_df["start_datetime"]).min()

                if "end_datetime" in inspection_df.columns.values:
                    inspection_end = pd.to_datetime(inspection_df["end_datetime"]).max()

                if inspection_start and inspection_end:
                    inspection_duration = inspection_end - inspection_start

            elif c_name in unashable_columns.index.values:
                if c_name == "hwinfo__cpu_features__flags":
                    initial_set = set(inspection_df[c_name][0])
                    difference = False
                    for flags_counter in range(1, len(inspection_df[c_name])):
                        if initial_set - set(inspection_df[c_name][flags_counter]):
                            difference = True

                    if not difference:
                        new_data[c_name] = [inspection_df[c_name][0]]
                    else:
                        new_data[c_name] = np.nan

                else:
                    values_column = inspection_df[c_name].apply(str).value_counts()
                    _LOGGER.debug(f"Skipped unashable column {c_name}: {values_column}")
                    new_data[c_name] = np.nan
            else:
                if len(inspection_df[c_name].unique()) == 1:
                    new_data[c_name] = [inspection_df[c_name].iloc[0]]
                else:
                    _LOGGER.debug(f"Skipped multiple values column: {c_name}")
                    new_data[c_name] = np.nan

        if inspection_duration:
            inspection_duration = inspection_duration.seconds

        new_data["inspection_start"] = [inspection_start]
        new_data["inspection_end"] = [inspection_end]
        new_data["inspection_duration"] = [inspection_duration]
        new_data["inspection_batch"] = [inspection_df.shape[0]]

        columns = [c for c in inspection_df.columns.values] + extra_columns
        return pd.DataFrame(new_data, columns=columns)

    @classmethod
    def create_inspections_dataframe(
        cls,
        processed_inspection_runs: Dict[str, pd.DataFrame],
        include_statistics: bool = False,
        performance_values: List[str] = ["elapsed_time", "rate"],
        parameter_for_statistics: str = "elapsed_time",
    ) -> pd.DataFrame:
        """Create final pd.DataFrame from processed inspections runs after evaluating statistics.

        :param processed_inspection_runs: dict with inspection results per inspection ID
        provided by `process_inspection_runs`.
        :param include_statistics: bool to decide if statistics are included in the dataframe
        :param parameter_for_statistics: parameter on which statistics are applied
        (it is used only when include_statistics=True)
        """
        if not processed_inspection_runs:
            _LOGGER.warning("No inspections runs have been received, no analysis can be performed.")
            return pd.DataFrame()

        if any(p_check not in cls._INSPECTION_PERFORMANCE_VALUES for p_check in performance_values):
            raise Exception(
                f"Performance parameters selected {performance_values}"
                f" are not all registered: {cls._INSPECTION_PERFORMANCE_VALUES}",
            )

        row_number = 0
        extracted_columns = []
        flags_columns = []

        for dataframe in processed_inspection_runs.values():

            for column in dataframe.columns.values:
                if column not in extracted_columns:
                    extracted_columns.append(column)

            for flags in dataframe["hwinfo__cpu_features__flags"].values:
                for flag in flags:
                    if f"flag__{flag}" not in flags_columns:
                        flags_columns.append(f"flag__{flag}")

        for flag_column in flags_columns:
            extracted_columns.append(flag_column)

        extra_columns = ["inspection_start", "inspection_end", "inspection_duration", "inspection_batch"]
        for extra_column in extra_columns:
            extracted_columns.append(extra_column)

        main_inspection_df = pd.DataFrame(columns=extracted_columns)

        column_names = [
            cls._INSPECTION_PERFORMANCE_VALUES[p_value] for p_value in performance_values
        ] + cls._INSPECTION_USAGE_VALUES

        for dataframe in processed_inspection_runs.values():

            new_df = cls.evaluate_statistics_on_inspection_df(
                inspection_df=dataframe,
                column_names=column_names,
                extra_columns=extra_columns,
            )

            for flags_ in dataframe["hwinfo__cpu_features__flags"].values:
                for flag_ in flags_:
                    new_df[f"flag__{flag_}"] = True

            main_inspection_df.loc[row_number] = new_df.iloc[0]
            row_number += 1

        if include_statistics:
            inspections_statistics_dataframe = AmunInspectionsStatistics.create_inspections_statistics_dataframe(
                processed_inspection_runs=processed_inspection_runs,
                parameters=[parameter_for_statistics],
            )
            return pd.merge(main_inspection_df, inspections_statistics_dataframe, on="inspection_document_id")

        return main_inspection_df

    @staticmethod
    def create_python_package_df(
        inspections_df: pd.DataFrame,
    ) -> Tuple[pd.DataFrame, Dict[str, Any], pd.DataFrame]:
        """Create DataFrame with only python packages present in software stacks.

        :param inspection_df: df of inspections results provided by `create_inspections_dataframe`.
        """
        python_packages_versions: Dict[str, Any] = {}
        python_packages_versions_plot: Dict[str, Any] = {}
        python_packages_names = []

        sws_df = inspections_df[[col for col in inspections_df.columns.values if "__index" in col]]

        for c_name in sws_df.columns.values:
            if "__index" in c_name:
                python_packages_names.append(c_name.split("__")[2])

        columns_packages = []
        for package in python_packages_names:
            columns_packages.append("".join(["requirements_locked__default__", package, "__index"]))
            columns_packages.append("".join(["requirements_locked__default__", package, "__version"]))

        for index, row in inspections_df[columns_packages].iterrows():

            for package in python_packages_names:
                version = row["".join(["requirements_locked__default__", package, "__version"])]
                index = row["".join(["requirements_locked__default__", package, "__index"])]

                if pd.isnull(version):
                    if package not in python_packages_versions.keys():
                        python_packages_versions[package] = []
                        python_packages_versions_plot[package] = []

                    python_packages_versions[package].append("")
                    python_packages_versions_plot[package].append("")

                else:
                    if package not in python_packages_versions.keys():
                        python_packages_versions[package] = []
                        python_packages_versions_plot[package] = []

                    python_packages_versions[package].append(f"{package}-{version.replace('==', '')}-{index}")
                    python_packages_versions_plot[package].append(version.replace("==", ""))

        return (
            pd.DataFrame(python_packages_versions),
            python_packages_versions,
            pd.DataFrame(python_packages_versions_plot),
        )

    @classmethod
    def create_final_dataframe(
        cls,
        inspections_df: pd.DataFrame,
        filters_for_identifiers: Optional[List[str]] = None,
        include_statistics: bool = False,
        use_only_versions: bool = False,
    ) -> pd.DataFrame:
        """Create final dataframe with all information required for plots.

        :param inspection_df: df of inspections results provided by `create_inspections_dataframe`.
        :param filters_for_identifiers: list of words to standardize identifiers.
        """
        if inspections_df.empty:
            _LOGGER.exception("Inspections dataframe is empty!")
            return pd.DataFrame()

        python_packages_dataframe, packages_versions_indexes, versions = cls.create_python_package_df(
            inspections_df=inspections_df,
        )

        label_encoder = LabelEncoder()

        if not use_only_versions:
            processed_string_result = copy.deepcopy(packages_versions_indexes)
        else:
            processed_string_result = copy.deepcopy(versions)

        sws_encoded = []
        for index, row in python_packages_dataframe.iterrows():
            sws_string = "<br>".join(["".join(pkg) for pkg in row.values if pkg != ""])
            hash_object = hashlib.sha256(bytes(sws_string, "raw_unicode_escape"))
            hex_dig = hash_object.hexdigest()
            sws_encoded.append([row.values, sws_string, hex_dig])

        re_encoded = []

        for index, row in inspections_df[
            ["os_release__id", "os_release__version_id", "requirements_locked___meta__requires__python_version"]
        ].iterrows():
            re_values = [re for re in row.values]
            re_values[2] = "".join(["py", "".join(re_values[2].split("."))])
            re_string = "-".join(re_values)
            hash_object = hashlib.sha256(bytes(re_string, "raw_unicode_escape"))
            hex_dig = hash_object.hexdigest()
            re_encoded.append([row.values, re_string, hex_dig])

        # Software Stack encoding
        processed_string_result["packages_list"] = [pp[0] for pp in sws_encoded]
        processed_string_result["sws_string"] = [pp[1] for pp in sws_encoded]
        processed_string_result["sws_hash_id"] = [pp[2] for pp in sws_encoded]

        sws_hash_id_values = array([pp[2] for pp in sws_encoded])

        integer_sws_hash_id_values_encoded = label_encoder.fit_transform(sws_hash_id_values)
        processed_string_result["sws_hash_id_encoded"] = integer_sws_hash_id_values_encoded

        # Runtime Environment:
        # Solver: OSName-OSVersion-PythonInterpreterVersion
        processed_string_result["solver"] = [pp[0] for pp in re_encoded]

        processed_string_result["os_name"] = [solver[0] for solver in processed_string_result["solver"]]
        processed_string_result["os_version"] = [solver[1] for solver in processed_string_result["solver"]]
        processed_string_result["python_interpreter"] = [solver[2] for solver in processed_string_result["solver"]]
        processed_string_result["solver_string"] = [pp[1] for pp in re_encoded]
        processed_string_result["solver_hash_id"] = [pp[2] for pp in re_encoded]

        processed_string_result["base"] = [
            cpu_family[0] for cpu_family in inspections_df[["specification_base"]].values
        ]

        # Hardware:
        # CPU
        processed_string_result["cpu_brand"] = [
            cpu_brand[0] for cpu_brand in inspections_df[["hwinfo__cpu_info__brand_raw"]].values
        ]
        processed_string_result["cpu_family"] = [
            cpu_family[0] for cpu_family in inspections_df[["runtime_environment__hardware__cpu_family"]].values
        ]
        processed_string_result["cpu_model"] = [
            cpu_model[0] for cpu_model in inspections_df[["runtime_environment__hardware__cpu_model"]].values
        ]
        processed_string_result["number_cpus"] = [
            number_cpus[0] for number_cpus in inspections_df[["run__requests__cpu"]].values
        ]
        # GPU
        processed_string_result["cuda_version"] = [
            cuda_version[0] for cuda_version in inspections_df[["runtime_environment__cuda_version"]].values
        ]

        # PI
        processed_string_result["pi_name"] = [pi_n[0] for pi_n in inspections_df[["stdout__name"]].values]
        processed_string_result["pi_component"] = [pi_c[0] for pi_c in inspections_df[["stdout__component"]].values]
        processed_string_result["pi_sha256"] = [pi_c[0] for pi_c in inspections_df[["script_sha256"]].values]

        # PI performance results
        processed_string_result["elapsed_time"] = [
            r_e[0] for r_e in inspections_df[["stdout__@result__elapsed"]].values
        ]
        processed_string_result["rate"] = [r_r[0] for r_r in inspections_df[["stdout__@result__rate"]].values]

        processed_string_result["inspection_document_id"] = [
            i[0] for i in inspections_df[["inspection_document_id"]].values
        ]

        final_df = pd.DataFrame(processed_string_result)

        standardized_identifiers = []

        if not filters_for_identifiers:
            filters_for_identifiers = []

        for _, row in inspections_df[["inspection_document_id", "identifier"]].iterrows():
            inspection_document_id = row["inspection_document_id"]
            identifier = row["identifier"]

            selected_identifer_ = identifier

            if not identifier:
                selected_identifer_ = inspection_document_id

            identifier_filter = "-".join(
                [word for word in selected_identifer_.split("-") if word not in filters_for_identifiers],
            )

            standardized_identifiers.append(identifier_filter)

        final_df["identifier"] = inspections_df["identifier"]
        final_df["standardized_identifier"] = standardized_identifiers

        final_df["start_datetime"] = inspections_df["inspection_start"]
        final_df["end_datetime"] = inspections_df["inspection_end"]

        final_df["total_duration"] = inspections_df["inspection_duration"]
        final_df["inspection_batch"] = inspections_df["inspection_batch"]

        if include_statistics:
            final_df["statistical_parameter"] = inspections_df["statistical_parameter"]
            final_df["std"] = inspections_df["std"]
            final_df["std_error"] = inspections_df["std_error"]
            final_df["mean"] = inspections_df["mean"]
            final_df["median"] = inspections_df["median"]
            final_df["q1"] = inspections_df["q1"]
            final_df["q3"] = inspections_df["q3"]
            final_df["iqr"] = inspections_df["iqr"]
            final_df["min"] = inspections_df["min"]
            final_df["max"] = inspections_df["max"]
            final_df["cov"] = inspections_df["cov"]
            final_df["covm"] = inspections_df["covm"]
            final_df["skew"] = inspections_df["skew"]

        return final_df

    @staticmethod
    def _convert_package_list_to_string(packages_version: List[str]) -> str:
        """Convert package list to str."""
        package_query = "["
        package_n = 1

        for package_name in packages_version:
            if package_n == 1:
                package_query += f'"{package_name}"'
            else:
                package_query += f', "{package_name}"'
            package_n += 1
        package_query += "]"

        return package_query

    @classmethod
    def filter_final_inspections_dataframe(
        cls,
        final_inspections_df: pd.DataFrame,
        inspection_document_ids: Optional[List[str]] = None,
        r_inspection_document_ids: Optional[List[str]] = None,
        standardized_ids: Optional[List[str]] = None,
        pi_name: Optional[List[str]] = None,
        pi_component: Optional[List[str]] = None,
        base: Optional[List[str]] = None,
        os_name: Optional[List[str]] = None,
        os_version: Optional[List[str]] = None,
        python_interpreter: Optional[List[str]] = None,
        cpu_family: Optional[List[str]] = None,
        cpu_model: Optional[List[str]] = None,
        cpus_number: Optional[List[str]] = None,
        packages: Optional[Dict[str, Any]] = None,
    ) -> pd.DataFrame:
        """Filter final inspections dataframe for plots.

        :param final_inspections_df: df for plots provided by `create_final_dataframe` or its subset.
        :param inspection_document_ids: fiter by inspection ids
        :param r_inspection_document_ids: remove by inspection ids
        :param standardized_ids: filter by standardized ids
        :param pi_name: fiter by performance indicator names (e.g PIMatmul)
        :param pi_component: filter by performance indicator components (e.g. tensorflow)
        :param base: filter by base images used e.g. quay.io/thoth-station/s2i-thoth-ubi8-py36
        :param os_name: filter by Operating System names e.g rhel
        :param os_version: filter by Operatin System versions e.g 8
        :param python_interpreter: filter by Python interpreters e.g 3.6
        :param cpu_family: filter by CPU family e.g 6
        :param cpu_model: filter by CPU model e.g. 85
        :param cpus_number: filter by number of CPUs e.g. 2
        :param packages: filter by packages in software stack: for each package {"name": ["name-version-index"]}.
        """
        if not final_inspections_df.shape[0]:
            _LOGGER.info("DataFrame provided is empty, nothing can be filtered.")

        filtered_df = final_inspections_df.copy()
        # Inspection IDs
        if inspection_document_ids:
            filtered_df.query("`inspection_document_id` == @inspection_document_ids", inplace=True)

        if r_inspection_document_ids:
            filtered_df.query("`inspection_document_id` != @r_inspection_document_ids", inplace=True)

        if standardized_ids:
            filtered_df.query("`standardized_identifier` == @standardized_ids", inplace=True)

        # Software stack
        if packages:
            counter = 1
            for package in packages:
                packages_version = packages[package]
                package_query = cls._convert_package_list_to_string(packages_version=packages_version)
                if counter == 1:
                    dynamic_query = f"`{package}` == {package_query}"
                else:
                    dynamic_query += f" and `{package}` == {package_query}"

                counter += 1
            filtered_df.query(dynamic_query, inplace=True)

        # Runtime Environment
        if base:
            filtered_df.query("base == @base", inplace=True)

        # Operating System
        if os_name:
            filtered_df.query("os_name == @os_name", inplace=True)

        if os_version:
            filtered_df.query("os_version == @os_version", inplace=True)

        # Python Interpreter
        if python_interpreter:
            filtered_df.query("python_interpreter == @python_interpreter", inplace=True)

        # Hardware
        if cpu_family:
            filtered_df.query("cpu_family == @cpu_family", inplace=True)

        if cpu_model:
            filtered_df.query("cpu_model == @cpu_model", inplace=True)

        if cpus_number:
            filtered_df.query("number_cpus == @cpus_number", inplace=True)

        # Performance Indicator (PI)
        if pi_name:
            filtered_df.query("pi_name == @pi_name", inplace=True)

        if pi_component:
            filtered_df.query("pi_component == @pi_component", inplace=True)

        if not filtered_df.shape[0]:
            _LOGGER.info("There are no results for the filters selected. Please change filters.")

        _LOGGER.info(f"Number of software stacks identified: {filtered_df.shape[0]}")

        return filtered_df

    @staticmethod
    def create_plot_results_summary(
        final_inspections_df: pd.DataFrame,
        performance_packages: List[str],
        include_statistics: bool = False,
    ) -> pd.DataFrame:
        """Show performance results summary.

        :param final_inspections_df: df for plots provided by `create_final_dataframe`.
        :param performance_packages: list of packages names
        :param include_statistics: include statistics in summary
        """
        identifier = ["inspection_document_id", "identifier", "standardized_identifier"]
        solver = ["os_name", "os_version", "base", "python_interpreter"]
        hardware = ["cpu_brand", "cpu_family", "cpu_model", "number_cpus"]
        runtime_environment = solver + hardware

        pi_info = ["pi_name"] + ["elapsed_time", "rate"]

        if include_statistics:
            statistics = [
                "statistical_parameter",
                "std",
                "std_error",
                "mean",
                "median",
                "q1",
                "q3",
                "iqr",
                "max",
                "min",
                "cov",
                "covm",
                "skew",
            ]
            return final_inspections_df[identifier + performance_packages + runtime_environment + pi_info + statistics]

        return final_inspections_df[identifier + performance_packages + runtime_environment + pi_info]


class AmunInspectionsStatistics:
    """Class of methods used to create statistics from Amun Inspections Runs."""

    _INSPECTION_MAPPING_PARAMETERS = {
        "elapsed_time": "stdout__@result__elapsed",
        "rate": "stdout__@result__rate",
        "utime": "usage__ru_utime",
        "stime": "usage__ru_stime",
        "nvcsw": "usage__ru_nvcsw",
        "nivcsw": "usage__ru_nivcsw",
    }

    @classmethod
    def _create_inspection_parameters_dataframe(
        cls,
        inspections_df: pd.DataFrame,
        parameters: List[str],
    ) -> pd.DataFrame:
        """Create pd.DataFrame of selected parameters from inspections_df.

        :param inspections_df: single inspection results  dataframe
        taken by `AmunInspections.process_inspection_runs` dictionary.
        :param parameters: inspection parameters used in the statistical analysis
        """
        renamed_columns = {cls._INSPECTION_MAPPING_PARAMETERS[parameter]: parameter for parameter in parameters}
        renamed_columns["stdout__name"] = "pi_name"

        filters = ["inspection_number", "inspection_document_id", "stdout__name"] + [
            cls._INSPECTION_MAPPING_PARAMETERS[parameter] for parameter in parameters
        ]
        subset_df = inspections_df[filters]
        subset_df.rename(columns=renamed_columns, inplace=True)

        return subset_df

    @classmethod
    def create_inspections_statistics_dataframe(
        cls,
        processed_inspection_runs: Dict[str, Any],
        parameters: List[str],
    ) -> pd.DataFrame:
        """Evaluate statistical quantities of each parameter selected for inspection results.

        :param processed_inspection_runs: dict with inspection results per inspection ID
        provided by `AmunInspections.process_inspection_runs`.
        :param parameters: inspection parameters used in the statistical analysis
        """
        results: List[Dict[str, Any]] = []

        for inspection_document_id in processed_inspection_runs:
            inspection_id_results_df = processed_inspection_runs[inspection_document_id]

            inspection_parameters_df = cls._create_inspection_parameters_dataframe(
                inspections_df=inspection_id_results_df,
                parameters=parameters,
            )

            for inspection_parameter in parameters:
                std = inspection_parameters_df[inspection_parameter].std()
                std_error = std / np.sqrt(inspection_parameters_df[inspection_parameter].shape[0])
                mean = inspection_parameters_df[inspection_parameter].mean()
                median = inspection_parameters_df[inspection_parameter].median()
                q1 = inspection_parameters_df[inspection_parameter].quantile([0.25])
                q3 = inspection_parameters_df[inspection_parameter].quantile([0.75])
                q1 = q1.iloc[[0.25]].values[0]
                q3 = q3.iloc[[0.75]].values[0]
                iqr = q3 - q1
                maxr = inspection_parameters_df[inspection_parameter].max()
                minr = inspection_parameters_df[inspection_parameter].min()
                cov = std / mean
                covm = iqr / median
                skew = inspection_parameters_df[inspection_parameter].skew()

                results.append(
                    {
                        "inspection_document_id": inspection_document_id,
                        "pi_name": inspection_parameters_df["pi_name"].unique()[0],
                        "statistical_parameter": inspection_parameter,
                        "std": std,
                        "std_error": std_error,
                        "mean": mean,
                        "median": median,
                        "q1": q1,
                        "q3": q3,
                        "iqr": iqr,
                        "max": maxr,
                        "min": minr,
                        "cov": cov,
                        "covm": covm,
                        "skew": skew,
                    },
                )

        inspections_statistics_dataframe = pd.DataFrame(results)

        return inspections_statistics_dataframe


class AmunInspectionsFailedSummary:
    """Class of methods used to compare failed with successfull inspections from Amun Inspections Runs."""

    @staticmethod
    def show_software_stack_differences(
        inspections_df: pd.DataFrame,
        failed_inspections_df: pd.DataFrame,
    ) -> pd.DataFrame:
        """Create summary report of the difference in the layers identified.

        :param inspection_df: df of inspections results provided by `AmunInspections.create_inspections_dataframe`.
        :param failed_inspections_df: df of failed inspections results
        provided by `AmunInspections.create_inspections_dataframe`.
        """
        results: List[Dict[str, Any]] = []
        if inspections_df.empty or failed_inspections_df.empty:
            _LOGGER.warning("No inspections runs have been received, no failures summary can be produced.")
            return pd.DataFrame(results)

        python_packages_dataframe, _, _ = AmunInspections.create_python_package_df(inspections_df=inspections_df)
        packages = set(python_packages_dataframe.columns.values)

        python_packages_dataframe_failed, _, _ = AmunInspections.create_python_package_df(
            inspections_df=failed_inspections_df,
        )
        packages_from_failed = set(python_packages_dataframe_failed.columns.values)

        total_packages = packages.union(packages_from_failed)

        for package in sorted(list(total_packages)):

            evaluated = 0

            success: Iterable[Any] = set()
            if package in python_packages_dataframe.columns.values:
                success = python_packages_dataframe[package].unique()
                evaluated += 1

            failed: Iterable[Any] = set()
            if package in python_packages_dataframe.columns.values:
                failed = python_packages_dataframe_failed[package].unique()
                evaluated += 1

            if evaluated == 2:
                difference: Iterable[Any] = set(failed) - set(success)
                common = set(failed) & set(success)

            results.append(
                {
                    "package": package,
                    "versions_in_successfull": sorted(success),
                    "versions_in_failed": sorted(failed),
                    "versions_in_failed_only": sorted(difference),
                    "versions_common": sorted(common),
                },
            )

        return pd.DataFrame(results)


class AmunInspectionsSummary:
    """Class of methods used to create summary from Amun Inspections Runs."""

    _INSPECTION_REPORT_FEATURES = {
        "hardware": {"title": "Hardware", "values": ["platform", "processor", "flags", "ncpus", "info"]},
        "base_image": {"title": "Operating System", "values": ["base_image", "number_cpus_run"]},
        "software_stack": {"title": "Software Stack", "values": ["requirements_locked"]},
        "pi": {"title": "Performance Indicator", "values": ["pi"]},
        "exit_codes": {"title": "Exit Code", "values": ["exit_code"]},
    }

    _INSPECTION_JSON_DF_KEYS_FEATURES_MAPPING = {
        "platform": {"description": "Platform", "values": ["hwinfo__platform"]},
        "processor": {"description": "Processor", "values": ["cpu_type__is", "cpu_type__has"]},
        "flags": {"description": "Flags", "values": ["flag__"]},
        "ncpus": {"description": "Number of CPUs", "values": ["hwinfo__cpu_type__ncpus"]},
        "info": {
            "description": "General info",
            "values": [
                "runtime_environment__hardware__cpu_family",
                "runtime_environment__hardware__cpu_model",
                "hwinfo__cpu_info__brand_raw",
                "runtime_environment__cuda_version",
            ],
        },
        "requirements_locked": {
            "description": "Packages",
            "values": ["requirements_locked__default", "requirements_locked___meta"],
        },
        "base_image": {
            "description": "Base Image",
            "values": ["os_release__name", "os_release__version", "specification_base"],
        },
        "number_cpus_run": {"description": "CPUs during run", "values": ["run__requests__cpu"]},
        "pi": {
            "description": "",
            "values": ["script_sha256", "@parameters", "stdout__name", "stdout__component", "batch_size"],
        },
        "exit_code": {"description": "", "values": ["exit_code"]},
    }

    @classmethod
    def _discover_unique_values(cls, objects: pd.DataFrame) -> pd.DataFrame:
        """Discover unique objects per context."""
        from deepdiff import DeepDiff  # For Deep Difference of 2 objects

        unique_objects = []
        first = objects.iloc[0].to_dict()
        unique_objects.append(objects.iloc[0].to_dict())

        for number in range(1, objects.shape[0]):
            new = objects.iloc[number].to_dict()
            ddiff = DeepDiff(first, new, ignore_order=True)

            if ddiff:
                unique = True
                for obj in unique_objects:
                    sddiff = DeepDiff(obj, new, ignore_order=True)
                    if not sddiff:
                        unique = False

                if unique:
                    unique_objects.append(new)

        return pd.DataFrame(unique_objects)

    @classmethod
    def produce_summary_report(
        cls,
        inspections_df: pd.DataFrame,
        is_markdown: Optional[bool] = False,
    ) -> Tuple[Dict[str, Any], Optional[str]]:
        """Create summary report of the difference in the layers identified.

        :param inspection_df: df of inspections results provided by `AmunInspections.create_inspections_dataframe`.
        """
        report_results: Dict[str, Any] = {}
        md_report_complete = ""

        if inspections_df.empty:
            _LOGGER.warning("No inspections runs have been received, no report summary can be produced.")
            return report_results, md_report_complete

        for feature in cls._INSPECTION_REPORT_FEATURES:
            md_report_complete += f"\n\n {cls._INSPECTION_REPORT_FEATURES[feature]['title']}"
            report_results[feature] = {}

            for report_part in cls._INSPECTION_REPORT_FEATURES[feature]["values"]:
                md_report_complete += (
                    f"\n\n {cls._INSPECTION_JSON_DF_KEYS_FEATURES_MAPPING[report_part]['description']}"
                )
                cols = cls._INSPECTION_JSON_DF_KEYS_FEATURES_MAPPING[report_part]["values"]
                extracted = inspections_df[[col for col in inspections_df.columns if any(c in col for c in cols)]]
                extracted = extracted.fillna("nan")
                unique_extracted = cls._discover_unique_values(extracted)

                if feature == "software_stack":
                    unique_extracted = unique_extracted[
                        [
                            col
                            for col in unique_extracted.columns.values
                            if any(s in col for s in ["__version", "__index", "__meta"])
                        ]
                    ]
                report_results[feature][report_part] = unique_extracted
                md_report_complete += "\n\n" + unique_extracted.transpose().to_markdown()

        if not is_markdown:
            return report_results, md_report_complete

        return report_results, md_report_complete



#!/usr/bin/env python3
# thoth-report-processing
# Copyright(C) 2020 Francesco Murdaca
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Security Indicators reports processing methods."""

import os
import logging
import json

from pathlib import Path
from typing import List, Optional, Tuple, Dict, Any, Union

import pandas as pd

from thoth.report_processing.exceptions import ThothSIPackageNotMatchingException
from thoth.report_processing.exceptions import ThothNotKnownResultStore
from thoth.report_processing.exceptions import ThothMissingDatasetAtPath
from thoth.report_processing.enums import ThothSecurityIndicatorsFileStoreEnum

from thoth.storages.security_indicators import SecurityIndicatorsResultsStore
from thoth.storages.security_indicators import SIAggregatedStore, SIClocStore, SIBanditStore

# set up logging
DEBUG_LEVEL = bool(int(os.getenv("DEBUG_LEVEL", 0)))

if DEBUG_LEVEL:
    logging.basicConfig(level=logging.DEBUG)
else:
    logging.basicConfig(level=logging.INFO)

_LOGGER = logging.getLogger(__name__)


class _SecurityIndicators:
    """Class of methods used to process reports from Security Indicators (SI) analyzer."""

    RESULTS_STORE = SecurityIndicatorsResultsStore

    def aggregate_thoth_security_indicators_results(
        self,
        store_files: Optional[List[str]] = None,
        limit_results: bool = False,
        max_ids: int = 5,
        is_local: bool = False,
        repo_path: Optional[Path] = None,
    ) -> Dict[str, Any]:
        """Aggregate results stored on Ceph or locally from repo for Thoth components reports.

        :param store_files: files to be retrieved from the Store for each result, if None all files are retrieved.
        :param limit_results: reduce the number of reports ids considered to `max_ids`.
        :param max_ids: maximum number of reports ids considered.
        :param is_local: flag to retrieve the dataset locally (if not uses Ceph S3 (credentials are required)).
        :param repo_path: required if you want to retrieve the dataset locally and `is_local` is set to True.
        """
        if store_files:
            if any(store_file not in ThothSecurityIndicatorsFileStoreEnum.__members__ for store_file in store_files):
                raise ThothNotKnownResultStore(
                    f"SecurityIndicatorsStore does not contain some of the files listed: {store_files}. \
                        \nSecurityIndicatorsStore: {ThothSecurityIndicatorsFileStoreEnum.__members__.keys()}",
                )

        if limit_results:
            _LOGGER.debug(f"Limiting results to {max_ids}!")

        files: Dict[str, Any] = {}

        if is_local:
            files, counter = self._aggregate_thoth_results_from_local(
                repo_path=repo_path,
                files=files,
                limit_results=limit_results,
                max_ids=max_ids,
                store_files=store_files,
            )

        else:
            files, counter = self._aggregate_thoth_results_from_ceph(
                store_files=store_files,
                files=files,
                limit_results=limit_results,
                max_ids=max_ids,
            )

        _LOGGER.info("Number of files retrieved is: %r" % counter)

        return files

    @staticmethod
    def _aggregate_thoth_results_from_local(
        files: Dict[str, Any],
        store_files: Optional[List[str]] = None,
        repo_path: Optional[Path] = None,
        limit_results: bool = False,
        max_ids: int = 5,
        is_multiple: Optional[bool] = None,
    ) -> Tuple[Dict[str, Any], int]:
        """Aggregate Thoth results from local repo."""
        _LOGGER.info(f"Retrieving dataset at path... {repo_path}")
        if not repo_path:
            _LOGGER.warning("No Path has been provided to retrieve data locally.")
            return files, 0

        if not repo_path.exists():
            raise ThothMissingDatasetAtPath(f"There is no dataset at this path: {repo_path}.")

        counter = 0

        for result_path in repo_path.iterdir():
            _LOGGER.info(f"Considering... {result_path}")

            if "security-indicator" not in result_path.name:
                raise Exception(f"This repo is not part of security-indicator! {result_path}")

            retrieved_files: Dict[str, Any] = {result_path.name: {}}

            for file_path in result_path.iterdir():

                if store_files and file_path.name in store_files:
                    with open(file_path, "r") as json_file_type:
                        json_file = json.load(json_file_type)

                    retrieved_files[result_path.name][file_path.name] = json_file

            files[result_path.name] = retrieved_files[result_path.name]

            counter += 1

            if limit_results:
                if counter == max_ids:
                    return files, counter

        return files, counter

    def _aggregate_thoth_results_from_ceph(
        self,
        files: Dict[str, Any],
        store_files: Optional[List[str]] = None,
        limit_results: bool = False,
        max_ids: int = 5,
    ) -> Tuple[Dict[str, Any], int]:
        """Aggregate Thoth results from Ceph."""
        store_class_type = self.RESULTS_STORE

        counter = 0

        files_id = []
        for security_indicator_key in store_class_type.iter_security_indicators():
            security_indicator_id = security_indicator_key.split("/")[0]

            if security_indicator_id not in files_id:
                _LOGGER.info(f"Document id: {security_indicator_id}")

                files_id.append(security_indicator_id)

                try:
                    retrieved_files: Dict[str, Any] = {security_indicator_id: {}}

                    if store_files and ThothSecurityIndicatorsFileStoreEnum.bandit.name in store_files:
                        si_bandit_store = SIBanditStore(security_indicator_id=security_indicator_id)
                        si_bandit_store.connect()

                        si_bandit_report = si_bandit_store.retrieve_document()
                        retrieved_files[security_indicator_id][
                            ThothSecurityIndicatorsFileStoreEnum.bandit.name
                        ] = si_bandit_report

                    if store_files and ThothSecurityIndicatorsFileStoreEnum.cloc.name in store_files:
                        si_cloc_store = SIClocStore(security_indicator_id=security_indicator_id)
                        si_cloc_store.connect()

                        si_cloc_report = si_cloc_store.retrieve_document()
                        retrieved_files[security_indicator_id][
                            ThothSecurityIndicatorsFileStoreEnum.cloc.name
                        ] = si_cloc_report

                    if store_files and ThothSecurityIndicatorsFileStoreEnum.aggregated.name in store_files:
                        si_aggregated_store = SIAggregatedStore(security_indicator_id=security_indicator_id)
                        si_aggregated_store.connect()

                        si_aggregated_report = si_aggregated_store.retrieve_document()
                        retrieved_files[security_indicator_id][
                            ThothSecurityIndicatorsFileStoreEnum.aggregated.name
                        ] = si_aggregated_report

                    files[security_indicator_id] = retrieved_files[security_indicator_id]

                    counter += 1

                    _LOGGER.info("Documents retrieved: %r", counter)

                    if limit_results:
                        if counter == max_ids:
                            return files, counter
                except Exception as si_exception:
                    _LOGGER.exception(
                        f"Exception during retrieval of SI result {security_indicator_id}: {si_exception}",
                    )
                    pass

        return files, counter


class SecurityIndicatorsBandit(_SecurityIndicators):
    """Class of methods used to process reports from Security Indicators (SI) bandit analyzer."""

    # Weights for Confidence
    HIGH_CONFIDENCE_WEIGHT = 1
    MEDIUM_CONFIDENCE_WEIGHT = 0.5
    LOW_CONFIDENCE_WEIGHT = 0.1

    # Weights for Security
    HIGH_SEVERITY_WEIGHT = 100
    MEDIUM_SEVERITY_WEIGHT = 10
    LOW_SEVERITY_WEIGHT = 1

    @staticmethod
    def aggregate_security_indicator_bandit_results(
        limit_results: bool = False,
        max_ids: int = 5,
        is_local: bool = False,
        security_indicator_bandit_repo_path: Path = Path("security-indicators"),
    ) -> List[Any]:
        """Aggregate si_bandit results from Ceph or locally using the provided path.

        :param limit_results: reduce the number of si_bandit reports ids considered to `max_ids` to test analysis.
        :param max_ids: maximum number of si_bandit reports ids considered.
        :param is_local: flag to retreive the dataset locally or from S3 (credentials are required).
        :param si_bandit_repo_path: path to retrieve the si_bandit dataset locally and `is_local` is set to True.
        """
        document_name = "bandit"
        si_reports: Dict[str, Any] = _SecurityIndicators().aggregate_thoth_security_indicators_results(
            store_files=[document_name],
            limit_results=limit_results,
            max_ids=max_ids,
            is_local=is_local,
            repo_path=security_indicator_bandit_repo_path,
        )
        security_indicator_bandit_reports = [
            document_results[document_name]
            for document_id, document_results in si_reports.items()
            if document_name in document_results
        ]

        _LOGGER.info("Number of files that can be used is: %r" % len(security_indicator_bandit_reports))

        return security_indicator_bandit_reports

    @staticmethod
    def _extract_data_from_si_bandit_metadata(si_bandit_report: Dict[str, Any]) -> Dict[str, Any]:
        """Extract data from si-bandit report metadata.

        :param si_bandit_report: SI bandit report provided by Thoth SI bandit analyzer.

        :output extracted_metadata: dictinary with metadata retrieved from SI bandit report.
        """
        report_metadata = si_bandit_report["metadata"]

        extracted_metadata = {
            "datetime_si_bandit": report_metadata["datetime"],
            "analyzer_si_bandit": report_metadata["analyzer"],
            "analyzer_version_si_bandit": report_metadata["analyzer_version"],
            "document_id_si_bandit": report_metadata["document_id"],
            "package_name": report_metadata["arguments"]["app.py"]["package_name"],
            "package_version": report_metadata["arguments"]["app.py"]["package_version"],
            "package_index": report_metadata["arguments"]["app.py"]["package_index"],
        }

        return extracted_metadata

    def create_si_bandit_metadata_dataframe(
        self,
        si_bandit_report: Dict[str, Any],
        analyzer_version: Optional[str] = None,
    ) -> pd.DataFrame:
        """Create si-bandit report metadata dataframe.

        :param si_bandit_report: SI bandit report provided by Thoth SI bandit analyzer.
        :param analyzer_version: analyzer version filter

        :output metadata_df: pandas.DataFrame with metadata obtained from `extract_data_from_si_bandit_metadata`.
        """
        if analyzer_version:
            document_id = si_bandit_report["metadata"]["document_id"]
            version = si_bandit_report["metadata"]["analyzer_version"]

            if not int("".join(version.split("."))) >= int("".join(analyzer_version.split("."))):
                _LOGGER.info(f"Skipping SI-bandit report: {document_id} because has version: {version}")
                return pd.DataFrame()

        metadata_si_bandit = self._extract_data_from_si_bandit_metadata(si_bandit_report=si_bandit_report)
        metadata_df = pd.DataFrame([metadata_si_bandit])

        return metadata_df

    @staticmethod
    def extract_severity_confidence_info(
        si_bandit_report: Dict[str, Any],
        filters_files: Optional[List[str]] = None,
    ) -> Tuple[List[Dict[str, Any]], Dict[str, int]]:
        """Extract severity and confidence from result metrics.

        :param si_bandit_report: SI bandit report provided by Thoth SI bandit analyzer.
        :param filters_files: List of strings of files to be filtered from analysis
        e.g. filter_files = ['/tests'] where /tests is filtered in the file path.

        :output extracted_info: list of dictionary for each SEVERITY/CONFIDENCE combination.
        :output summary_files: dictionary with statistics about analyzed, filtered total files.
        """
        extracted_info: List[Dict[str, Any]] = []

        summary_files = {
            "number_of_analyzed_files": 0,
            "number_of_files_with_severities": 0,
            "number_of_filtered_files": 0,
        }

        if not filters_files:
            filters_files = []

        si_bandit_result = si_bandit_report["result"]

        si_bandit_report_result_metrics_df = pd.DataFrame(si_bandit_result["metrics"])
        si_bandit_report_result_results_df = pd.DataFrame(si_bandit_result["results"])

        if "filename" not in si_bandit_report_result_results_df.columns.values:
            return extracted_info, summary_files

        for file in si_bandit_report_result_metrics_df.columns.values:
            # Filter tests/ file
            if file != "_totals" and not any(filter_ in file for filter_ in filters_files):

                analysis = {}
                analysis["name"] = file

                analysis["SEVERITY.LOW"] = {
                    "CONFIDENCE.LOW": 0,
                    "CONFIDENCE.MEDIUM": 0,
                    "CONFIDENCE.HIGH": 0,
                    "CONFIDENCE.UNDEFINED": 0,
                }
                analysis["SEVERITY.MEDIUM"] = {
                    "CONFIDENCE.LOW": 0,
                    "CONFIDENCE.MEDIUM": 0,
                    "CONFIDENCE.HIGH": 0,
                    "CONFIDENCE.UNDEFINED": 0,
                }
                analysis["SEVERITY.HIGH"] = {
                    "CONFIDENCE.LOW": 0,
                    "CONFIDENCE.MEDIUM": 0,
                    "CONFIDENCE.HIGH": 0,
                    "CONFIDENCE.UNDEFINED": 0,
                }

                subset_df = si_bandit_report_result_results_df[
                    si_bandit_report_result_results_df["filename"].values == file
                ]
                if subset_df.shape[0] > 0:
                    # check if there are severities for the file

                    for index, row in subset_df[["issue_confidence", "issue_severity"]].iterrows():
                        analysis[f"SEVERITY.{row['issue_confidence']}"][f"CONFIDENCE.{row['issue_severity']}"] += 1

                    summary_files["number_of_files_with_severities"] += 1

                summary_files["number_of_analyzed_files"] += 1

                extracted_info.append(analysis)

            elif file != "_totals" and any(filter_ in file for filter_ in filters_files):
                summary_files["number_of_filtered_files"] += 1

        return extracted_info, summary_files

    def create_security_confidence_dataframe(
        self,
        si_bandit_report: Dict[str, Any],
        filters_files: Optional[List[str]] = None,
    ) -> Tuple[pd.DataFrame, Dict[str, int]]:
        """Create Security/Confidence dataframe for si-bandit report.

        :param si_bandit_report: SI bandit report provided by Thoth SI bandit analyzer.
        :param filters_files: List of strings of files to be filtered from analysis
        e.g. filter_files = ['/tests'] where /tests is filtered in the file path.

        :output sec_conf_df: pandas.DataFrame with all info about SEVERITY/CONFIDENCE for the package analyzed.
        :output summary_files: dictionary with statistics about analyzed, filtered total files.
        """
        results_sec_conf, summary_files = self.extract_severity_confidence_info(
            si_bandit_report=si_bandit_report,
            filters_files=filters_files,
        )

        summary_df = pd.DataFrame()

        if results_sec_conf:
            summary_df = pd.json_normalize(results_sec_conf, sep="__").set_index("name")
        else:
            summary_df = pd.json_normalize(results_sec_conf, sep="__")

        summary_df["_total_severity"] = summary_df.sum(axis=1)
        sec_conf_df = summary_df.transpose()
        sec_conf_df["_total"] = sec_conf_df.sum(axis=1)

        return sec_conf_df, summary_files

    @staticmethod
    def produce_si_bandit_report_summary_dataframe(
        metadata_df: pd.DataFrame,
        si_bandit_sec_conf_df: pd.DataFrame,
        summary_files: Dict[str, int],
    ) -> pd.DataFrame:
        """Create si-bandit report summary dataframe.

        :param metadata_df: pandas.DataFrame provided by `create_si_bandit_metadata_dataframe`.
        :param sec_conf_df: pandas.DataFrame provided by `create_security_confidence_dataframe`.
        :output summary_files: dictionary with statistics about analyzed, filtered total files.

        :output report_summary_df: pandas.DataFrame summary for a single si bandit report.
        """
        subset_df = pd.DataFrame([si_bandit_sec_conf_df["_total"].to_dict()])
        report_summary_df = pd.concat([metadata_df, subset_df], axis=1)
        report_summary_df["number_of_files_with_severities"] = pd.to_numeric(
            summary_files["number_of_files_with_severities"],
        )
        report_summary_df["number_of_analyzed_files"] = pd.to_numeric(summary_files["number_of_analyzed_files"])
        report_summary_df["number_of_filtered_files"] = pd.to_numeric(summary_files["number_of_filtered_files"])
        report_summary_df["number_of_files_total"] = (
            pd.to_numeric(
                summary_files["number_of_filtered_files"],
            )
            + pd.to_numeric(summary_files["number_of_analyzed_files"])
        )
        report_summary_df["_total_severity"] = pd.to_numeric(report_summary_df["_total_severity"])

        return report_summary_df

    def create_si_bandit_final_dataframe(
        self,
        si_bandit_report: Dict[str, Any],
        filters_files: Optional[List[str]] = None,
        analyzer_version: Optional[str] = None,
    ) -> pd.DataFrame:
        """Create final si-bandit dataframe.

        :param si_bandit_report: SI bandit report provided by Thoth SI bandit analyzer.
        :param filters_files: List of strings of files to be filtered from analysis
        e.g. filter_files = ['/tests'] where /tests is filtered in the file path.
        :param analyzer_version: analyzer version filter

        :output report_summary_df: pandas.DataFrame summary for a single si bandit report.
        """
        if analyzer_version:
            document_id = si_bandit_report["metadata"]["document_id"]
            version = si_bandit_report["metadata"]["analyzer_version"]

            if not int("".join(version.split("."))) >= int("".join(analyzer_version.split("."))):
                _LOGGER.info(f"Skipping SI-bandit report: {document_id} because has version: {version}")
                return pd.DataFrame()

        # Create metadata dataframe
        metadata_df = self.create_si_bandit_metadata_dataframe(
            si_bandit_report=si_bandit_report,
            analyzer_version=analyzer_version,
        )

        if analyzer_version and metadata_df.empty:
            _LOGGER.info(f"Skipping SI-bandit report: {document_id} because has version: {version}")
            return pd.DataFrame()

        # Create metadata dataframe
        package_name = metadata_df["package_name"][0]
        package_version = metadata_df["package_version"][0]
        package_index = metadata_df["package_index"][0]

        _LOGGER.info(f"Analyzing si_bandit report for package_name: {package_name}")
        _LOGGER.info(f"Analyzing si_bandit report for package_version: {package_version}")
        _LOGGER.info(f"Analyzing si_bandit report for package_index: {package_index}")

        # Create Security/Confidence dataframe
        security_confidence_df, summary_files = self.create_security_confidence_dataframe(
            si_bandit_report=si_bandit_report,
            filters_files=filters_files,
        )

        # Create Summary dataframe
        si_bandit_report_summary_df = self.produce_si_bandit_report_summary_dataframe(
            metadata_df=metadata_df,
            si_bandit_sec_conf_df=security_confidence_df,
            summary_files=summary_files,
        )

        return si_bandit_report_summary_df

    def aggregate_si_bandit_final_dataframe(
        self,
        si_bandit_reports: List[Dict[str, Any]],
        filters_files: Optional[List[str]] = None,
        analyzer_version: Optional[str] = None,
    ) -> pd.DataFrame:
        """Aggregate si-bandit dataframes into final dataframe.

        :param si_bandit_reports: list of SI bandit report provided by Thoth SI bandit analyzer.
        :param filters_files: List of strings of files to be filtered from analysis
        e.g. filter_files = ['/tests'] where /tests is filtered in the file path.
        :param analyzer_version: analyzer version filter

        :output final_df: pandas.DataFrame aggregating all SI bandit reports provided.
        """
        counter = 1
        final_df = pd.DataFrame()
        total_reports = len(si_bandit_reports)

        for si_bandit_report in si_bandit_reports:

            document_id = si_bandit_report["metadata"]["document_id"]
            _LOGGER.info(f"Analyzing SI-bandit report: {counter}/{total_reports}")

            si_bandit_report_summary_df = self.create_si_bandit_final_dataframe(
                si_bandit_report=si_bandit_report,
                filters_files=filters_files,
                analyzer_version=analyzer_version,
            )
            if not si_bandit_report_summary_df.empty:
                final_df = pd.concat([final_df, si_bandit_report_summary_df], axis=0)

                counter += 1
            else:
                _LOGGER.info(f"Skipping SI-bandit report: {document_id} because has different version")

        final_df.reset_index(inplace=True, drop=True)

        return final_df

    def create_security_indicators_scores(self, si_bandit_df: pd.DataFrame) -> pd.DataFrame:
        """Create Security Indicators (SI) scores from si bandit outputs.

        :param si_bandit_df: pandas.DataFrame as given by `aggregate_si_bandit_final_dataframe`.

        :output si_bandit_df: Extend `si_bandit_df` with SI scores created using all rows (aka all packages).
        """
        if not any("SEVERITY" in column for column in si_bandit_df.columns):
            si_bandit_df["SEVERITY.score"] = 0
            si_bandit_df["SEVERITY.score.normalized"] = 0
            _LOGGER.exception("All reports considered have no vulnerabilities")
            return si_bandit_df

        for security in ["LOW", "MEDIUM", "HIGH"]:
            for confidence in ["LOW", "MEDIUM", "HIGH"]:

                vulnerability_class = f"SEVERITY.{security}__CONFIDENCE.{confidence}"

                min_max_scaler = (si_bandit_df[vulnerability_class] - si_bandit_df[vulnerability_class].min()) / (
                    si_bandit_df[vulnerability_class].max() - si_bandit_df[vulnerability_class].min()
                )

                si_bandit_df[f"{vulnerability_class}_scaled"] = min_max_scaler

        si_bandit_df["SEVERITY.HIGH.sub_score"] = (
            si_bandit_df["SEVERITY.HIGH__CONFIDENCE.HIGH_scaled"].fillna(0) * self.HIGH_CONFIDENCE_WEIGHT
            + si_bandit_df["SEVERITY.HIGH__CONFIDENCE.MEDIUM_scaled"].fillna(0) * self.MEDIUM_CONFIDENCE_WEIGHT
            + si_bandit_df["SEVERITY.HIGH__CONFIDENCE.LOW_scaled"].fillna(0) * self.LOW_CONFIDENCE_WEIGHT
        ) / 3

        si_bandit_df["SEVERITY.MEDIUM.sub_score"] = (
            si_bandit_df["SEVERITY.MEDIUM__CONFIDENCE.HIGH_scaled"].fillna(0) * self.HIGH_CONFIDENCE_WEIGHT
            + si_bandit_df["SEVERITY.MEDIUM__CONFIDENCE.MEDIUM_scaled"].fillna(0) * self.MEDIUM_CONFIDENCE_WEIGHT
            + si_bandit_df["SEVERITY.MEDIUM__CONFIDENCE.LOW_scaled"].fillna(0) * self.LOW_CONFIDENCE_WEIGHT
        ) / 3

        si_bandit_df["SEVERITY.LOW.sub_score"] = (
            si_bandit_df["SEVERITY.LOW__CONFIDENCE.HIGH_scaled"].fillna(0) * self.HIGH_CONFIDENCE_WEIGHT
            + si_bandit_df["SEVERITY.LOW__CONFIDENCE.MEDIUM_scaled"].fillna(0) * self.MEDIUM_CONFIDENCE_WEIGHT
            + si_bandit_df["SEVERITY.LOW__CONFIDENCE.LOW_scaled"].fillna(0) * self.LOW_CONFIDENCE_WEIGHT
        ) / 3

        si_bandit_df["SEVERITY.score"] = (
            si_bandit_df["SEVERITY.HIGH.sub_score"] * self.HIGH_SEVERITY_WEIGHT
            + si_bandit_df["SEVERITY.MEDIUM.sub_score"] * self.MEDIUM_SEVERITY_WEIGHT
            + si_bandit_df["SEVERITY.LOW.sub_score"] * self.LOW_SEVERITY_WEIGHT
        ) / 3

        si_bandit_df["SEVERITY.score.normalized"] = (
            si_bandit_df["SEVERITY.score"] / si_bandit_df["number_of_analyzed_files"].max()
        )

        return si_bandit_df


class SecurityIndicatorsCloc:
    """Class of methods used to process reports from Security Indicators (SI) cloc analyzer."""

    @staticmethod
    def aggregate_security_indicator_cloc_results(
        limit_results: bool = False,
        max_ids: int = 5,
        is_local: bool = False,
        security_indicator_cloc_repo_path: Path = Path("security-indicators"),
    ) -> List[Any]:
        """Aggregate si_cloc results from Ceph or locally using the provided path.

        :param limit_results: reduce the number of si_cloc reports ids considered to `max_ids` to test analysis
        :param max_ids: maximum number of si_cloc reports ids considered
        :param is_local: flag to retreive the dataset locally or from S3 (credentials are required)
        :param si_cloc_repo_path: path to retrieve the si_cloc dataset locally and `is_local` is set to True
        """
        document_name = "cloc"

        si_reports: Dict[str, Any] = _SecurityIndicators().aggregate_thoth_security_indicators_results(
            store_files=[document_name],
            limit_results=limit_results,
            max_ids=max_ids,
            is_local=is_local,
            repo_path=security_indicator_cloc_repo_path,
        )
        security_indicator_cloc_reports = [
            document_results[document_name]
            for document_id, document_results in si_reports.items()
            if document_name in document_results
        ]

        _LOGGER.info("Number of files that can be used is: %r" % len(security_indicator_cloc_reports))

        return security_indicator_cloc_reports

    @staticmethod
    def _extract_data_from_si_cloc_metadata(si_cloc_report: Dict[str, Any]) -> Dict[str, Any]:
        """Extract data from si-cloc report metadata.

        :param si_cloc_report: SI cloc report provided by Thoth SI cloc analyzer.
        :output extracted_metadata: dictinary with metadata retrieved from SI cloc report.
        """
        report_metadata = si_cloc_report["metadata"]

        extracted_metadata = {
            "datetime_si_cloc": report_metadata["datetime"],
            "analyzer_si_cloc": report_metadata["analyzer"],
            "analyzer_version_si_cloc": report_metadata["analyzer_version"],
            "document_id_si_cloc": report_metadata["document_id"],
            "package_name": report_metadata["arguments"]["si-cloc"]["package_name"],
            "package_version": report_metadata["arguments"]["si-cloc"]["package_version"],
            "package_index": report_metadata["arguments"]["si-cloc"]["package_index"],
        }

        return extracted_metadata

    def create_si_cloc_metadata_dataframe(
        self,
        si_cloc_report: Dict[str, Any],
        analyzer_version: Optional[str] = None,
    ) -> pd.DataFrame:
        """Create si-cloc report metadata dataframe.

        :param si_cloc_report: SI cloc report provided by Thoth SI cloc analyzer.
        :param analyzer_version: analyzer version filter

        :output metadata_df: pandas.DataFrame with metadata obtained from `extract_data_from_si_cloc_metadata`.
        """
        if analyzer_version:
            document_id = si_cloc_report["metadata"]["document_id"]
            version = si_cloc_report["metadata"]["analyzer_version"]

            if not int("".join(version.split("."))) >= int("".join(analyzer_version.split("."))):
                _LOGGER.info(f"Skipping SI-cloc report: {document_id} because has version: {version}")
                return pd.DataFrame()

        metadata_si_cloc = self._extract_data_from_si_cloc_metadata(si_cloc_report=si_cloc_report)
        metadata_df = pd.DataFrame([metadata_si_cloc])

        return metadata_df

    def create_si_cloc_results_dataframe(self, si_cloc_report: Dict[str, Any]) -> pd.DataFrame:
        """Create si-cloc report results dataframe."""
        results = {k: v for k, v in si_cloc_report["result"].items() if k != "header"}
        results["SUM"]["n_lines"] = si_cloc_report["result"]["header"]["n_lines"]
        results_df = pd.json_normalize(results)

        return results_df

    @staticmethod
    def produce_si_cloc_report_summary_dataframe(
        metadata_df: pd.DataFrame,
        cloc_results_df: pd.DataFrame,
    ) -> pd.DataFrame:
        """Create si-cloc report summary dataframe.

        :param metadata_df: pandas.DataFrame provided by `create_si_cloc_metadata_dataframe`.
        :param cloc_results_df: pandas.DataFrame provided by `create_si_cloc_results_dataframe`.

        :output report_summary_df: pandas.DataFrame summary for a single si cloc report.
        """
        report_summary_df = pd.concat([metadata_df, cloc_results_df], axis=1)

        return report_summary_df

    def create_si_cloc_final_dataframe(
        self,
        si_cloc_report: Dict[str, Any],
        analyzer_version: Optional[str] = None,
    ) -> pd.DataFrame:
        """Create final si-cloc final dataframe.

        :param si_cloc_report: SI cloc report provided by Thoth SI cloc analyzer.
        :param analyzer_version: analyzer version filter

        :output report_summary_df: pandas.DataFrame summary for a single si cloc report.
        """
        if analyzer_version:
            document_id = si_cloc_report["metadata"]["document_id"]
            version = si_cloc_report["metadata"]["analyzer_version"]

            if not int("".join(version.split("."))) >= int("".join(analyzer_version.split("."))):
                _LOGGER.info(f"Skipping SI-cloc report: {document_id} because has version: {version}")
                return pd.DataFrame()

        # Create metadata dataframe
        metadata_df = self.create_si_cloc_metadata_dataframe(si_cloc_report, analyzer_version=analyzer_version)

        if analyzer_version and metadata_df.empty:
            _LOGGER.info(f"Skipping SI-cloc report: {document_id} because has version: {version}")
            return pd.DataFrame()

        package_name = metadata_df["package_name"][0]
        package_version = metadata_df["package_version"][0]
        package_index = metadata_df["package_index"][0]

        _LOGGER.info(f"Analyzing si_cloc report for package_name: {package_name}")
        _LOGGER.info(f"Analyzing si_cloc report for package_version: {package_version}")
        _LOGGER.info(f"Analyzing si_cloc report for package_index: {package_index}")

        # Create cloc results dataframe
        cloc_results_df = self.create_si_cloc_results_dataframe(si_cloc_report=si_cloc_report)

        report_summary_df = self.produce_si_cloc_report_summary_dataframe(
            metadata_df=metadata_df,
            cloc_results_df=cloc_results_df,
        )

        return report_summary_df

    def aggregate_si_cloc_final_dataframes(
        self,
        si_cloc_reports: List[Dict[str, Any]],
        analyzer_version: Optional[str] = None,
    ) -> pd.DataFrame:
        """Aggregate si-cloc dataframes into final dataframe.

        :param si_cloc_reports: list of SI cloc report provided by Thoth SI cloc analyzer.
        :param analyzer_version: analyzer version filter

        :output final_df: pandas.DataFrame aggregating all SI cloc reports provided.
        """
        counter = 1
        total_reports = len(si_cloc_reports)

        final_df = pd.DataFrame()

        for si_cloc_report in si_cloc_reports:

            document_id = si_cloc_report["metadata"]["document_id"]
            _LOGGER.info(f"Analyzing SI-cloc report: {counter}/{total_reports}")

            si_cloc_report_summary_df = self.create_si_cloc_final_dataframe(
                si_cloc_report=si_cloc_report,
                analyzer_version=analyzer_version,
            )

            if not si_cloc_report_summary_df.empty:
                final_df = pd.concat([final_df, si_cloc_report_summary_df], axis=0)

                counter += 1
            else:
                _LOGGER.info(f"Skipping SI-cloc report: {document_id} because has different version")

        final_df.reset_index(inplace=True, drop=True)

        return final_df


class SecurityIndicatorsAggregator:
    """Class of methods used to aggregate reports from Security Indicators (SI) analyzers."""

    si_bandit = SecurityIndicatorsBandit()
    si_cloc = SecurityIndicatorsCloc()

    @staticmethod
    def retrieve_security_indicator_aggregated_results(
        limit_results: bool = False,
        max_ids: int = 5,
        is_local: bool = False,
        security_indicator_aggregated_repo_path: Path = Path("security-indicators"),
    ) -> List[Any]:
        """Retrieve si_aggregated results from Ceph or locally using the provided path.

        :param limit_results: reduce the number of si_aggregated reports ids considered to `max_ids` to test analysis.
        :param max_ids: maximum number of si_aggregated reports ids considered.
        :param is_local: flag to retreive the dataset locally or from S3 (credentials are required).
        :param si_aggregated_repo_path: path to retrieve the si_aggregated data locally and `is_local` is set to True.
        """
        document_name = "aggregated"

        si_reports: Dict[str, Any] = _SecurityIndicators().aggregate_thoth_security_indicators_results(
            store_files=[document_name],
            limit_results=limit_results,
            max_ids=max_ids,
            is_local=is_local,
            repo_path=security_indicator_aggregated_repo_path,
        )
        security_indicator_aggregated_reports = [
            document_results[document_name]
            for document_id, document_results in si_reports.items()
            if document_name in document_results
        ]

        _LOGGER.info("Number of files that can be used is: %r" % len(security_indicator_aggregated_reports))

        return security_indicator_aggregated_reports

    def create_si_aggregated_results(
        self,
        si_bandit_report: Dict[str, Any],
        si_cloc_report: Dict[str, Any],
        filters_files: Optional[List[str]] = None,
        si_bandit_version: Optional[str] = None,
        si_cloc_version: Optional[str] = None,
        output_json: bool = False,
    ) -> Union[pd.DataFrame, Dict[str, Any]]:
        """Create dataframe or json with aggregated data from SI analyzers.

        :param si_bandit_report: SI bandit report provided by Thoth SI bandit analyzer.
        :param si_cloc_report: SI cloc report provided by Thoth SI cloc analyzer.
        :param filters_files: List of strings of files to be filtered from analysis
        e.g. filter_files = ['/tests'] where /tests is filtered in the file path.
        :param si_bandit_version: filter for si bandit analyzer version
        :param si_cloc_version: filter for si cloc analyzer version
        :param output_json: if json output is required

        :output aggregated_df: pandas.DataFrame aggregating all SI analyzers reports provided.
        :output aggregated_json: pandas.DataFrame aggregating all SI analyzers reports provided.
        """
        aggregated_df = pd.DataFrame()
        si_bandit_df = self.si_bandit.create_si_bandit_final_dataframe(
            si_bandit_report=si_bandit_report,
            filters_files=filters_files,
            analyzer_version=si_bandit_version,
        )
        si_cloc_df = self.si_cloc.create_si_cloc_final_dataframe(
            si_cloc_report=si_cloc_report,
            analyzer_version=si_cloc_version,
        )
        if si_bandit_df.empty or si_cloc_df.empty:
            _LOGGER.exception("One of the analyzer results is empty!")

        package_info = ["package_name", "package_version", "package_index"]
        si_bandit_package = set(str(v) for v in si_bandit_df[package_info].values)
        si_cloc_package = set(str(v) for v in si_cloc_df[package_info].values)

        if si_bandit_package - si_cloc_package:
            raise ThothSIPackageNotMatchingException(
                "The reports are from different packages, cannot be aggregated!"
                f"\nsi_bandit:{si_bandit_package}"
                f"\nsi_cloc: {si_cloc_package}",
            )

        package_df = si_bandit_df[package_info]

        si_bandit_df.drop(columns=package_info, inplace=True)
        si_cloc_df.drop(columns=package_info, inplace=True)

        aggregated_df = pd.concat([package_df, si_bandit_df, si_cloc_df], axis=1)

        aggregated_df.reset_index(inplace=True, drop=True)

        if output_json:
            aggregated_si = aggregated_df.to_json(orient="records")  # string

            aggregated_json: Dict[str, Any] = json.loads(aggregated_si)[0]

            return aggregated_json

        return aggregated_df



#!/usr/bin/env python3
# thoth-report-processing
# Copyright(C) 2020 Francesco Murdaca, Sai Sankar Gochhayat
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Adviser reports processing methods."""

import logging
import os
import json

from datetime import datetime, date
from pathlib import Path
from typing import List, Optional, Dict, Any, Tuple

import pandas as pd
import numpy as np

from thoth.report_processing.exceptions import ThothMissingDatasetAtPath
from thoth.common import normalize_os_version
from thoth.common import map_os_name
from thoth.storages import CephStore
from thoth.storages.advisers import AdvisersResultsStore

# set up logging
DEBUG_LEVEL = bool(int(os.getenv("DEBUG_LEVEL", 0)))

if DEBUG_LEVEL:
    logging.basicConfig(level=logging.DEBUG)
else:
    logging.basicConfig(level=logging.INFO)

_LOGGER = logging.getLogger(__name__)


class Adviser:
    """Class of methods used to process results from Adviser."""

    @classmethod
    def aggregate_adviser_results(
        cls,
        start_date: Optional[date] = None,
        end_date: Optional[date] = None,
        limit_results: bool = False,
        max_ids: int = 5,
        is_local: bool = False,
        repo_path: Optional[Path] = None,
    ) -> Dict[str, Any]:
        """Aggregate results stored on Ceph or locally from repo for Thoth components reports.

        :param start_date: start date for documents to consider.
        :param end_date: end date for documents to consider.
        :param limit_results: reduce the number of reports ids considered to `max_ids`.
        :param max_ids: maximum number of reports ids considered.
        :param is_local: flag to retrieve the dataset locally (if not uses Ceph S3 (credentials are required)).
        :param repo_path: required if you want to retrieve the dataset locally and `is_local` is set to True.
        """
        if limit_results:
            _LOGGER.debug(f"Limiting results to {max_ids}!")

        files: Dict[str, Any] = {}

        if not is_local:
            files, counter = cls._aggregate_thoth_results_from_ceph(
                files=files,
                start_date=start_date,
                end_date=end_date,
                limit_results=limit_results,
                max_ids=max_ids,
            )
            _LOGGER.info("Number of files retrieved is: %r" % counter)

            return files

        files, counter = cls._aggregate_thoth_results_from_local(
            repo_path=repo_path,
            files=files,
            limit_results=limit_results,
            max_ids=max_ids,
        )
        _LOGGER.info("Number of files retrieved is: %r" % counter)

        return files

    @staticmethod
    def _aggregate_thoth_results_from_local(
        files: Dict[str, Any],
        repo_path: Optional[Path] = None,
        limit_results: bool = False,
        max_ids: int = 5,
    ) -> Tuple[Dict[str, Any], int]:
        """Aggregate Thoth results from local repo."""
        _LOGGER.info(f"Retrieving dataset at path... {repo_path}")
        if not repo_path:
            _LOGGER.warning("No Path has been provided to retrieve data locally.")
            return files, 0

        if not repo_path.exists():
            raise ThothMissingDatasetAtPath(f"There is no dataset at this path: {repo_path}.")

        counter = 0

        for file_path in repo_path.iterdir():
            _LOGGER.info(f"Considering... {file_path}")

            if "adviser" not in file_path.name:
                raise Exception(f"This repo is not part of adviser! {repo_path}")

            with open(file_path, "r") as json_file_type:
                json_file = json.load(json_file_type)

            files[file_path.name] = json_file

            counter += 1

            if limit_results:
                if counter == max_ids:
                    return files, counter

        return files, counter

    @staticmethod
    def _aggregate_thoth_results_from_ceph(
        files: Dict[str, Any],
        start_date: Optional[date] = None,
        end_date: Optional[date] = None,
        limit_results: bool = False,
        max_ids: int = 5,
    ) -> Tuple[Dict[str, Any], int]:
        """Aggregate Thoth results from Ceph."""
        adviser_store = AdvisersResultsStore()
        adviser_store.connect()

        number_adviser_results = adviser_store.get_document_count(start_date=start_date, end_date=end_date)

        _LOGGER.info("Number of Adviser reports identified is: %r" % number_adviser_results)

        document_ids = [idd for idd in adviser_store.get_document_listing(start_date=start_date, end_date=end_date)]

        counter = 0

        if limit_results:
            _LOGGER.debug(f"Limiting results to {max_ids} to test functions!!")

        for document_id in document_ids:
            _LOGGER.debug(f"Analysis {document_id} n.{counter + 1}/{number_adviser_results}")

            try:
                document = adviser_store.retrieve_document(document_id)

                if "metadata" in document.keys():
                    files[document_id] = document

                    counter += 1

                    _LOGGER.info("Documents retrieved: %r", counter)

                    if limit_results:
                        if counter == max_ids:
                            return files, counter
                else:
                    _LOGGER.warning(f"'metadata' is not present in {document_id} keys!")

            except Exception as exception:
                _LOGGER.exception(f"Exception during retrieval of adviser result {document_id}: {exception}")
                continue

        return files, counter

    @staticmethod
    def _update_statistics(
        statistics: Dict[str, Any],
        analyzer_version: str,
        error_count: int,
        no_report: bool = False,
    ) -> Dict[str, Any]:
        if analyzer_version not in statistics:
            statistics[analyzer_version] = {}

        if analyzer_version not in statistics[analyzer_version]:
            statistics[analyzer_version]["adviser_version"] = analyzer_version
        key = "success"

        if no_report or error_count:
            key = "failure"

        if key not in statistics[analyzer_version]:
            statistics[analyzer_version][key] = 1
        else:
            statistics[analyzer_version][key] += 1

        return statistics

    @classmethod
    def _retrieve_adviser_justifications(
        cls,
        adviser_files: Dict[str, Any],
    ) -> Tuple[List[Dict[str, Any]], Dict[str, Any]]:
        """Retrieve adviser justifications.

        :param adviser_files: adviser documents
        """
        statistics: Dict[str, Any] = {}

        justifications_collected: List[Dict[str, Any]] = []

        for document_id, document in adviser_files.items():

            error_count = 0
            report = {}

            try:
                datetime_advise_run = document["metadata"].get("datetime")
                analyzer_version = document["metadata"].get("analyzer_version")
                datetime_object = datetime.strptime(datetime_advise_run, "%Y-%m-%dT%H:%M:%S.%f")
                result = document["result"]
                report = result.get("report")

            except Exception as report_error:
                _LOGGER.error(f"Error analyzing adviser document {document_id} report stack info: {report_error}")
                _LOGGER.error("Adviser document %s report is: %s", document_id, report)

            if not report:
                _LOGGER.warning(f"No report for adviser document: {document_id}, checking error msg from document.")
                error_msg = result.get("error_msg", "no report provided")

                justifications_collected.append(
                    {
                        "document_id": document_id,
                        "date": datetime_object,
                        "analyzer_version": analyzer_version,
                        "justification": error_msg,
                        "error": True,
                        "message": error_msg,
                        "type": "ERROR",
                    },
                )

                cls._update_statistics(
                    statistics=statistics,
                    analyzer_version=analyzer_version,
                    error_count=error_count,
                    no_report=True,
                )

                continue

            try:
                justifications_collected, error_count = cls.extract_adviser_justifications_from_stack_info(
                    report=report,
                    justifications_collected=justifications_collected,
                    document_id=document_id,
                    datetime_object=datetime_object,
                    analyzer_version=analyzer_version,
                    error_count=error_count,
                )

            except Exception as stack_info_error:
                _LOGGER.error(f"Error analyzing adviser document {document_id} report stack info: {stack_info_error}")
                _LOGGER.error("Adviser document %s report stack info: %s", document_id, report.get("stack_info"))

            cls._update_statistics(
                statistics=statistics,
                analyzer_version=analyzer_version,
                error_count=error_count,
            )

            try:
                justifications_collected = cls.extract_adviser_justifications_from_products(
                    report=report,
                    justifications_collected=justifications_collected,
                    document_id=document_id,
                    datetime_object=datetime_object,
                    analyzer_version=analyzer_version,
                )
            except Exception as products_error:
                _LOGGER.error(f"Error analyzing adviser document {document_id} report products: {products_error}")
                _LOGGER.error("Adviser document %s report products: %s", document_id, report.get("products"))

        if statistics:
            if "success" not in statistics[analyzer_version]:
                statistics[analyzer_version]["success"] = 0

            if "failure" not in statistics[analyzer_version]:
                statistics[analyzer_version]["failure"] = 0

        return justifications_collected, statistics

    @classmethod
    def create_adviser_justifications_and_statistics_dataframe(
        cls,
        adviser_files: Dict[str, Any],
    ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """Create dataframe of adviser justifications from results."""
        justifications_collected, statistics = cls._retrieve_adviser_justifications(adviser_files=adviser_files)
        adviser_justifications_dataframe = pd.DataFrame(justifications_collected)

        if not adviser_justifications_dataframe.empty:
            adviser_justifications_dataframe["date_"] = [
                pd.to_datetime(str(v_date)).strftime("%Y-%m-%d")
                for v_date in adviser_justifications_dataframe["date"].values
            ]

        adviser_statistics_dataframe = pd.DataFrame(statistics)

        if not adviser_statistics_dataframe.empty:
            # Maintain order not to lose track when storing on csv without headers
            adviser_statistics_dataframe = adviser_statistics_dataframe.transpose()[
                ["adviser_version", "success", "failure"]
            ]
        return adviser_justifications_dataframe, adviser_statistics_dataframe

    @staticmethod
    def _retrieve_adviser_inputs_statistics(
        adviser_files: Dict[str, Any],
    ) -> List[Dict[str, Any]]:
        """Retrieve adviser inputs statistics.

        :param adviser_files: adviser documents
        """
        adviser_inputs_collected: List[Dict[str, Any]] = []

        for document_id, document in adviser_files.items():

            datetime_advise_run = document["metadata"].get("datetime")
            datetime_object = datetime.strptime(datetime_advise_run, "%Y-%m-%dT%H:%M:%S.%f")

            cli_arguments = document["metadata"]["arguments"]["thoth-adviser"]
            source_type = (cli_arguments.get("metadata") or {}).get("source_type")
            source_type = source_type.upper() if source_type else None

            parameters = document["result"]["parameters"]

            runtime_environment = parameters["project"].get("runtime_environment")
            os = runtime_environment.get("operating_system", {})
            if os:
                os_name = runtime_environment["operating_system"].get("name")
                if os_name:
                    runtime_environment["operating_system"]["name"] = map_os_name(
                        os_name=runtime_environment["operating_system"]["name"],
                    )

            # Recommendation type
            recommendation_type = parameters["recommendation_type"].upper()

            # Solver
            os = runtime_environment.get("operating_system", {})
            os_name = os.get("name")
            os_version = normalize_os_version(os.get("name"), os.get("version"))
            python_interpreter = runtime_environment.get("python_version")

            # Base image
            base_image = runtime_environment.get("base_image", None)
            # Hardware
            hardware = runtime_environment.get("hardware", {})
            adviser_inputs_collected.append(
                {
                    "document_id": document_id,
                    "date": datetime_object,
                    "source_type": source_type,
                    "recommendation_type": recommendation_type,
                    "base_image": base_image,
                    "solver": f'{os_name}-{os_version}-py{python_interpreter.replace(".", "")}',
                    "cpu_model": hardware.get("cpu_model", None),
                    "cpu_family": hardware.get("cpu_family", None),
                },
            )

        return adviser_inputs_collected

    @classmethod
    def create_adviser_inputs_info_dataframe(cls, adviser_files: Dict[str, Any]) -> pd.DataFrame:
        """Create dataframe of adviser inputs info from results."""
        adviser_inputs_info_dataframe = pd.DataFrame(
            cls._retrieve_adviser_inputs_statistics(adviser_files=adviser_files),
        )
        if not adviser_inputs_info_dataframe.empty:
            adviser_inputs_info_dataframe["date_"] = [
                pd.to_datetime(str(v_date)).strftime("%Y-%m-%d")
                for v_date in adviser_inputs_info_dataframe["date"].values
            ]
        return adviser_inputs_info_dataframe

    @classmethod
    def create_adviser_dataframes(cls, adviser_files: Dict[str, Any]) -> Dict[str, pd.DataFrame]:
        """Create dataframe of adviser justifications from results."""
        dataframes = {}
        (
            dataframes["justifications"],
            dataframes["statistics"],
        ) = cls.create_adviser_justifications_and_statistics_dataframe(adviser_files=adviser_files)
        dataframes["inputs_info"] = cls.create_adviser_inputs_info_dataframe(adviser_files=adviser_files)

        return dataframes

    @staticmethod
    def extract_adviser_justifications_from_stack_info(
        report: Dict[str, Any],
        justifications_collected: List[Dict[str, Any]],
        document_id: str,
        datetime_object: datetime,
        analyzer_version: str,
        error_count: int,
    ) -> Tuple[List[Dict[str, Any]], int]:
        """Retrieve justifications from stack info from adviser report."""
        stack_info = report.get("stack_info")

        if not stack_info:
            _LOGGER.warning("No stack info in report.")
            return justifications_collected, error_count

        for info in stack_info:

            justification = {"message": info["message"], "type": info["type"]}

            if "link" in info:
                message = info["link"]
            else:
                message = info["message"]

            error = False

            if info["type"] == "ERROR":
                error = True
                error_count += 1

            justifications_collected.append(
                {
                    "document_id": document_id,
                    "date": datetime_object,
                    "analyzer_version": analyzer_version,
                    "justification": justification,
                    "error": error,
                    "message": message,
                    "type": info["type"],
                },
            )

        return justifications_collected, error_count

    @classmethod
    def extract_adviser_justifications_from_products(
        cls,
        report: Dict[str, Any],
        justifications_collected: List[Dict[str, Any]],
        document_id: str,
        datetime_object: datetime,
        analyzer_version: str,
    ) -> List[Dict[str, Any]]:
        """Retrieve justifications from products from adviser report."""
        products = report.get("products")
        justifications_collected = cls.extract_justifications_from_products(
            products=products,
            justifications_collected=justifications_collected,
            document_id=document_id,
            datetime_object=datetime_object,
            analyzer_version=analyzer_version,
        )

        return justifications_collected

    @staticmethod
    def extract_justifications_from_products(
        products: Optional[List[Dict[str, Any]]],
        justifications_collected: List[Dict[str, Any]],
        document_id: str,
        datetime_object: datetime,
        analyzer_version: str,
    ) -> List[Dict[str, Any]]:
        """Extract justifications from products in adviser document."""
        if not products:
            _LOGGER.debug(f"No products identified in adviser document: {document_id}")
            return justifications_collected

        # TODO: Handle all products
        product = products[0]
        justifications = product["justification"]

        if justifications:
            # Collect all justifications
            for justification in justifications:

                if "advisory" in justification:

                    error = True
                    message = justification["advisory"]
                    justification_type = justification["type"]

                elif "link" in justification:

                    error = False
                    message = justification["link"]
                    justification_type = justification["type"]

                else:
                    error = False
                    message = justification["message"]
                    justification_type = justification["type"]

                justifications_collected.append(
                    {
                        "document_id": document_id,
                        "date": datetime_object,
                        "analyzer_version": analyzer_version,
                        "justification": justification,
                        "error": error,
                        "message": message,
                        "type": justification_type,
                    },
                )

        else:
            _LOGGER.warning(f"No justifications identified for adviser report: {document_id}")

        return justifications_collected

    @staticmethod
    def create_adviser_results_dataframe_histogram(adviser_type_dataframe: pd.DataFrame) -> pd.DataFrame:
        """Create adviser results dataframe sorted for histogram plot.

        :param adviser_type_dataframe dataframe as given by any of df outputs in `create_adviser_dataframe`
        """
        histogram_data: Dict[str, Any] = {}

        for i in adviser_type_dataframe[["message", "type"]].index:
            message = adviser_type_dataframe[["message", "type"]].loc[i, "message"]

            if message not in histogram_data.keys():
                histogram_data[message] = {
                    "message": message,
                    "type": adviser_type_dataframe[["message", "type"]].loc[i, "type"],
                    "count": adviser_type_dataframe["message"].value_counts()[message],
                }

        sorted_justifications_df = pd.DataFrame(histogram_data)
        sorted_justifications_df = sorted_justifications_df.transpose()
        sorted_justifications_df = sorted_justifications_df.sort_values(by="count", ascending=False)

        return sorted_justifications_df

    @staticmethod
    def _aggregate_data_per_interval(adviser_type_dataframe: pd.DataFrame, number_days: int = 7) -> pd.DataFrame:
        """Aggregate advise justifications per days intervals.

        :param adviser_type_dataframe dataframe produced by `create_adviser_dataframe`
        """
        begin = min(adviser_type_dataframe["date"].values)
        end = max(adviser_type_dataframe["date"].values)

        timestamps = []

        delta = np.timedelta64(number_days, "D")
        intervals = (end - begin) / delta
        value = begin

        for i in range(0, int(intervals) + 1):
            value = value + delta
            timestamps.append(value)

        timestamps[0] = begin
        timestamps[len(timestamps) - 1] = end

        aggregated_data: Dict[str, Any] = {}

        # Iterate over all intervals
        for interval in range(0, len(timestamps)):
            low = timestamps[interval - 1]
            high = timestamps[interval]
            aggregated_data[high] = {}
            subset_df = adviser_type_dataframe[
                (adviser_type_dataframe["date"] >= low) & (adviser_type_dataframe["date"] <= high)
            ]
            messages = pd.unique(subset_df["message"])

            for message in messages:
                if message not in aggregated_data[high].keys():
                    aggregated_data[high][message] = {
                        "message": message,
                        "count": subset_df["message"].value_counts()[message],
                    }
        return aggregated_data

    @staticmethod
    def _create_heatmaps_values(input_data: Dict[str, Any], advise_encoded_type: List[int]) -> Dict[str, Any]:
        """Create values for heatmaps."""
        heatmaps_values: Dict[str, Any] = {}

        for advise_type in set(advise_encoded_type):
            _LOGGER.debug(f"Analyzing advise type... {advise_type}")
            type_values = []

            for upper_interval, interval_runs in input_data.items():
                _LOGGER.debug(f"Checking for that advise type in 'interval'... {upper_interval}")

                if advise_type in interval_runs.keys():
                    type_values.append(interval_runs[advise_type]["count"])
                else:
                    type_values.append(0)

            heatmaps_values[str(advise_type)] = type_values

        return heatmaps_values

    @classmethod
    def create_adviser_results_dataframe_heatmap(
        cls,
        adviser_type_dataframe: pd.DataFrame,
        number_days: int = 7,
    ) -> pd.DataFrame:
        """Create adviser justifications heatmap.

        :param adviser_type_dataframe dataframe as given by any of df outputs in `create_adviser_dataframe`
        :param number_days: number of days to split data.
        """
        data = cls._aggregate_data_per_interval(adviser_type_dataframe=adviser_type_dataframe, number_days=number_days)
        heatmaps_values = cls._create_heatmaps_values(
            input_data=data,
            advise_encoded_type=adviser_type_dataframe["message"].values,
        )
        df_heatmap = pd.DataFrame(heatmaps_values)
        df_heatmap["interval"] = data.keys()
        df_heatmap = df_heatmap.set_index(["interval"])
        df_heatmap = df_heatmap.transpose()

        justifications_ordered = []
        for message in pd.unique(adviser_type_dataframe["message"]):
            justifications_ordered.append(message)

        df_heatmap["advise_type"] = justifications_ordered

        df_heatmap = df_heatmap.set_index(["advise_type"])

        return df_heatmap

    @staticmethod
    def _get_processed_data_prefix(ceph_bucket_prefix: str, processed_data_name: str, environment: str) -> str:
        """Get prefix where processed data are stored."""
        bucket_prefix = ceph_bucket_prefix
        deployment_name = os.environ["THOTH_DEPLOYMENT_NAME"]
        return f"{bucket_prefix}/{deployment_name}/{processed_data_name}-{environment}"

    @classmethod
    def connect_to_ceph(
        cls,
        ceph_bucket_prefix: str,
        processed_data_name: str,
        environment: str,
        bucket: Optional[str] = None,
    ) -> CephStore:
        """Connect to Ceph to store processed data."""
        prefix = cls._get_processed_data_prefix(
            ceph_bucket_prefix=ceph_bucket_prefix,
            processed_data_name=processed_data_name,
            environment=environment,
        )
        ceph = CephStore(prefix=prefix, bucket=bucket)
        ceph.connect()
        return ceph

    @staticmethod
    def store_csv_from_dataframe(
        csv_from_df: str,
        ceph_sli: CephStore,
        file_name: str,
        ceph_path: str,
        is_public: bool = False,
    ) -> None:
        """Store CSV obtained from pd.DataFrame on Ceph.

        param: csv_from_df: CSV given from pd.DataFrame.to_csv()
        """
        if is_public:
            _LOGGER.info(f"Storing on public bucket... {ceph_path}")
        else:
            _LOGGER.info(f"Storing on private bucket... {ceph_path}")
        ceph_sli.store_blob(blob=csv_from_df, object_key=ceph_path)
        _LOGGER.info(f"Succesfully stored  {file_name} at {ceph_path}")

    @staticmethod
    def create_pretty_report_from_json(report: Dict[str, Any], is_justification: bool = False) -> str:
        """Create Markdown output from adviser report input."""
        md_report_complete = ""
        if not report:
            return md_report_complete

        products = report.get("products")
        if not products:
            return md_report_complete

        md_report_complete = "Report"

        md_report_complete += "\n\n" + "Justifications"

        final_df = pd.DataFrame(columns=["message", "type"])

        counter = 0
        # Consider the first product
        product = products[0]

        if "justification" in product.keys():
            justifications = product["justification"]

            if justifications:

                for justification in justifications:
                    final_df.loc[counter] = pd.DataFrame([justification]).iloc[0]
                    counter += 1

        md_report_complete += "\n\n" + final_df.to_markdown()

        if is_justification:
            return md_report_complete

        # Packages in Advised Pipfile
        md_report_complete += Adviser._add_packages_in_advised_pipfile_to_md_report(product=product, is_dev=False)

        # Dev-Packages in Advised Pipfile
        md_report_complete += Adviser._add_packages_in_advised_pipfile_to_md_report(product=product, is_dev=True)

        requirements = product["project"]["requirements"]

        if "requires" in requirements:
            if requirements["requires"]:
                md_report_complete += "\n\n" + "Requires in Advised Pipfile"
                df = pd.DataFrame([requirements["requires"]])
                md_report_complete += "\n\n" + df.to_markdown()

        if "source" in requirements:
            if requirements["source"]:
                md_report_complete += "\n\n" + "Source in Advised Pipfile"
                df = pd.DataFrame(requirements["source"])
                md_report_complete += "\n\n" + df.to_markdown()

        # Packages in Advised Pipfile.lock
        md_report_complete += Adviser._add_packages_in_advised_pipfile_lock_to_md_report(product=product, is_dev=False)

        # Dev-Packages in Advised Pipfile.lock
        md_report_complete += Adviser._add_packages_in_advised_pipfile_lock_to_md_report(product=product, is_dev=True)

        # Runtime Environment
        md_report_complete += Adviser._add_runtime_environment_to_md_report(product=product)

        if "score" in product:
            if product["score"]:
                md_report_complete += "\n\n" + "Software Stack Score"
                df = pd.DataFrame([{"score": product["score"]}])
                md_report_complete += "\n\n" + df.to_markdown()

        return md_report_complete

    @staticmethod
    def _add_packages_in_advised_pipfile_to_md_report(product: Dict[str, Any], is_dev: bool) -> str:
        """Add Packages in Advised Pipfile to final report."""
        md_report = ""

        if is_dev:
            packages = "dev-packages"
        else:
            packages = "packages"

        if not product["project"]["requirements"][packages]:
            return md_report

        if is_dev:
            md_report += "\n\n" + "Dev-Packages in Advised Pipfile"
        else:
            md_report += "\n\n" + "Packages in Advised Pipfile"

        packages_names = []
        packages_versions = []

        for package_name, requested_version in product["project"]["requirements"][packages].items():
            packages_names.append(package_name)
            packages_versions.append(requested_version)

        data = {"package_name": packages_names, "package_version": packages_versions}

        df = pd.DataFrame(data)
        md_report += "\n\n" + df.to_markdown()

        return md_report

    @staticmethod
    def _add_packages_in_advised_pipfile_lock_to_md_report(product: Dict[str, Any], is_dev: bool) -> str:
        """Add Packages in Advised Pipfile.lock to final report."""
        md_report = ""

        if is_dev:
            packages = "develop"
        else:
            packages = "default"

        if not product["project"]["requirements_locked"][packages]:
            return md_report

        if is_dev:
            md_report += "\n\n" + "Dev-Packages in Advised Pipfile.lock"
        else:
            md_report += "\n\n" + "Packages in Advised Pipfile.lock"

            packages_names = []
            packages_versions = []
            packages_indexes = []

            for package_name, data in product["project"]["requirements_locked"][packages].items():
                packages_names.append(package_name)
                packages_versions.append(data["version"])
                packages_indexes.append(data["index"])

            data = {"package_name": packages_names, "package_version": packages_versions, "index": packages_indexes}
            df = pd.DataFrame(data)
            md_report += "\n\n" + df.to_markdown()

        return md_report

    @staticmethod
    def _add_runtime_environment_to_md_report(product: Dict[str, Any]) -> str:
        """Add Packages in Advised Pipfile.lock to final report."""
        md_report = ""

        if not product["project"]["runtime_environment"]:
            return md_report

        runtime_environment = product["project"]["runtime_environment"]
        md_report += "\n\n" + "Runtime Environment"

        if "name" in runtime_environment:
            if runtime_environment["name"]:
                md_report += "\n\n" + "Runtime Environment - Name"
                df = pd.DataFrame([{"name": runtime_environment["name"]}])
                md_report += "\n\n" + df.to_markdown()

        if "cuda_version" in runtime_environment:
            if runtime_environment["cuda_version"]:
                md_report += "\n\n" + "Runtime Environment - CUDA"
                df = pd.DataFrame([{"cuda_version": runtime_environment["cuda_version"]}])
                md_report += "\n\n" + df.to_markdown()

        if "hardware" in runtime_environment:
            if [h for h in runtime_environment["hardware"].keys() if runtime_environment["hardware"][h]]:
                md_report += "\n\n" + "Runtime Environment - Hardware"
                df = pd.DataFrame([runtime_environment["hardware"]])
                md_report += "\n\n" + df.to_markdown()

        return md_report



#!/usr/bin/env python3
# thoth-report-processing
# Copyright(C) 2020 Francesco Murdaca
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Security Indicator bandit test suite."""

from .base_test import ReportProcessingTestCase
from thoth.report_processing.components.security import SecurityIndicatorsBandit


class TestSecurityReportsBandit(ReportProcessingTestCase):
    """Test implementation of security indicator bandit."""

    _SI_REPORT_NAME = "security-indicator-54c6daf9"
    _SI_BANDIT_FOLDER_PATH = ReportProcessingTestCase.DATA / "security-indicator"

    def test_get_security_indicator_bandit_report(self) -> None:
        """Test retrieving report from local path."""
        si_bandit_reports = SecurityIndicatorsBandit.aggregate_security_indicator_bandit_results(
            security_indicator_bandit_repo_path=self._SI_BANDIT_FOLDER_PATH,
            is_local=True,
        )
        assert si_bandit_reports[0]

    def test_get_metadata_df_from_bandit_report(self) -> None:
        """Test obtaining metadata from si bandit report."""
        si_bandit_reports = SecurityIndicatorsBandit.aggregate_security_indicator_bandit_results(
            security_indicator_bandit_repo_path=self._SI_BANDIT_FOLDER_PATH,
            is_local=True,
        )
        si_bandit_report = si_bandit_reports[0]
        si_bandit = SecurityIndicatorsBandit()
        metadata_retrieved = si_bandit._extract_data_from_si_bandit_metadata(si_bandit_report=si_bandit_report)
        metadata_retrieved_keys = [k for k in metadata_retrieved]
        metadata_test_keys = [
            "datetime_si_bandit",
            "analyzer_si_bandit",
            "analyzer_version_si_bandit",
            "document_id_si_bandit",
            "package_name",
            "package_version",
            "package_index",
        ]
        assert metadata_retrieved_keys == metadata_test_keys

    def test_get_severity_confidence_info_from_bandit_report(self) -> None:
        """Test obtaining severity confidence info from si bandit report."""
        si_bandit_reports = SecurityIndicatorsBandit.aggregate_security_indicator_bandit_results(
            security_indicator_bandit_repo_path=self._SI_BANDIT_FOLDER_PATH,
            is_local=True,
        )
        si_bandit_report = si_bandit_reports[0]
        si_bandit = SecurityIndicatorsBandit()
        severity_confidence_info, summary = si_bandit.extract_severity_confidence_info(
            si_bandit_report=si_bandit_report,
        )
        assert not severity_confidence_info

    def test_get_severity_confidence_info_df_from_bandit_report(self) -> None:
        """Test obtaining severity confidence info DataFrame from si bandit report."""
        si_bandit_reports = SecurityIndicatorsBandit.aggregate_security_indicator_bandit_results(
            security_indicator_bandit_repo_path=self._SI_BANDIT_FOLDER_PATH,
            is_local=True,
        )
        si_bandit_report = si_bandit_reports[0]
        si_bandit = SecurityIndicatorsBandit()
        severity_confidence_info_df, summary = si_bandit.create_security_confidence_dataframe(
            si_bandit_report=si_bandit_report,
        )

        severity_confidence_info = severity_confidence_info_df["_total"].to_dict()

        severity_confidence_info_keys = [k for k in severity_confidence_info]
        severity_confidence_info_test_keys = [
            "_total_severity",
        ]
        assert severity_confidence_info_keys == severity_confidence_info_test_keys



#!/usr/bin/env python3
# thoth-report-processing
# Copyright(C) 2020 Francesco Murdaca
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Solver reports processing methods."""

import logging
import os
import json

from pathlib import Path
from typing import List, Optional, Dict, Any, Tuple

from thoth.report_processing.exceptions import ThothMissingDatasetAtPath

from thoth.storages.solvers import SolverResultsStore

# set up logging
DEBUG_LEVEL = bool(int(os.getenv("DEBUG_LEVEL", 0)))

if DEBUG_LEVEL:
    logging.basicConfig(level=logging.DEBUG)
else:
    logging.basicConfig(level=logging.INFO)

_LOGGER = logging.getLogger(__name__)


class Solver:
    """Class of methods used to process results from Solver."""

    @classmethod
    def aggregate_solver_results(
        cls,
        limit_results: bool = False,
        max_ids: int = 5,
        is_local: bool = False,
        repo_path: Optional[Path] = None,
    ) -> Dict[str, Any]:
        """Aggregate results stored on Ceph or locally from repo for Thoth components reports.

        :param limit_results: reduce the number of reports ids considered to `max_ids`.
        :param max_ids: maximum number of reports ids considered.
        :param is_local: flag to retrieve the dataset locally (if not uses Ceph S3 (credentials are required)).
        :param repo_path: required if you want to retrieve the dataset locally and `is_local` is set to True.
        """
        if limit_results:
            _LOGGER.debug(f"Limiting results to {max_ids}!")

        files: Dict[str, Any] = {}

        if not is_local:
            files, counter = cls._aggregate_thoth_results_from_ceph(
                files=files,
                limit_results=limit_results,
                max_ids=max_ids,
            )
            _LOGGER.info("Number of files retrieved is: %r" % counter)

            return files

        files, counter = cls._aggregate_thoth_results_from_local(
            repo_path=repo_path,
            files=files,
            limit_results=limit_results,
            max_ids=max_ids,
        )
        _LOGGER.info("Number of files retrieved is: %r" % counter)

        return files

    @staticmethod
    def _aggregate_thoth_results_from_local(
        files: Dict[str, Any],
        store_files: Optional[List[str]] = None,
        repo_path: Optional[Path] = None,
        limit_results: bool = False,
        max_ids: int = 5,
    ) -> Tuple[Dict[str, Any], int]:
        """Aggregate Thoth results from local repo."""
        _LOGGER.info(f"Retrieving dataset at path... {repo_path}")
        if not repo_path:
            _LOGGER.warning("No Path has been provided to retrieve data locally.")
            return files, 0

        if not repo_path.exists():
            raise ThothMissingDatasetAtPath(f"There is no dataset at this path: {repo_path}.")

        counter = 0

        for file_path in repo_path.iterdir():
            _LOGGER.info(f"Considering... {file_path}")

            if "solver" not in file_path.name:
                raise Exception(f"This repo is not part of solver! {repo_path}")

            with open(file_path, "r") as json_file_type:
                json_file = json.load(json_file_type)

            files[file_path.name] = json_file

            counter += 1

            if limit_results:
                if counter == max_ids:
                    return files, counter

        return files, counter

    @staticmethod
    def _aggregate_thoth_results_from_ceph(
        files: Dict[str, Any],
        store_files: Optional[List[str]] = None,
        limit_results: bool = False,
        max_ids: int = 5,
    ) -> Tuple[Dict[str, Any], int]:
        """Aggregate Thoth results from Ceph."""
        solver_store = SolverResultsStore()
        solver_store.connect()

        solver_ids = list(solver_store.get_document_listing())

        _LOGGER.info("Number of Solver reports identified is: %r" % len(solver_ids))

        number_solver_results = len(solver_ids)

        counter = 0

        if limit_results:
            _LOGGER.debug(f"Limiting results to {max_ids} to test functions!!")

        for n, document_id in enumerate(solver_ids):
            _LOGGER.debug(f"Analysis {document_id} n.{counter + 1}/{number_solver_results}")

            try:
                document = solver_store.retrieve_document(document_id)
                files[document_id] = document

                counter += 1

                _LOGGER.info("Documents retrieved: %r", counter)

                if limit_results:
                    if counter == max_ids:
                        return files, counter

            except Exception as exception:
                _LOGGER.exception(f"Exception during retrieval of solver result {document_id}: {exception}")
                continue

        return files, counter

    @staticmethod
    def construct_solver_from_metadata(solver_report_metadata: Dict[str, Any]) -> str:
        """Construct solver from solver report metadata."""
        os_name = solver_report_metadata["os_release"]["name"].lower()
        os_version = "".join(
            [
                release_version
                for release_version in solver_report_metadata["os_release"]["version"]
                if release_version.isdigit()
            ],
        )
        python_interpreter = f'{solver_report_metadata["python"]["major"]}{solver_report_metadata["python"]["minor"]}'
        solver = f"{os_name}-{os_version}-py{python_interpreter}"

        return solver

    @classmethod
    def extract_data_from_solver_metadata(cls, solver_report_metadata: Dict[str, Any]) -> Dict[str, Any]:
        """Extract data from solver report metadata."""
        solver = cls.construct_solver_from_metadata(solver_report_metadata)
        solver_parts = solver.split("-")

        requirements = solver_report_metadata["arguments"]["python"]["requirements"]

        extracted_metadata = {
            "document_id": solver_report_metadata["document_id"],
            "datetime": solver_report_metadata["datetime"],
            "requirements": requirements,
            "solver": solver,
            "os_name": solver_parts[0],
            "os_version": solver_parts[1],
            "python_interpreter": ".".join(solver_parts[2][2:]),
            "analyzer_version": solver_report_metadata["analyzer_version"],
        }

        return extracted_metadata

    @staticmethod
    def extract_tree_from_solver_result(solver_report_result: Dict[str, Any]) -> List[Dict[str, Any]]:
        """Extract data from solver report result."""
        packages = []
        for python_package_info in solver_report_result["tree"]:
            package = {
                "package_name": python_package_info["package_name"],
                "package_version": python_package_info["package_version_requested"],
                "index_url": python_package_info["index_url"],
                "importlib_metadata": python_package_info["importlib_metadata"]["metadata"],
                "dependencies": python_package_info["dependencies"],
            }
            packages.append(package)

        return packages

    @staticmethod
    def extract_errors_from_solver_result(solver_report_result_errors: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """Extract all errors from solver report (if any)."""
        errors = []
        for error in solver_report_result_errors:
            errors.append(
                {
                    "package_name": error["package_name"],
                    "package_version": error["package_version"],
                    "index_url": error["index_url"],
                    "type": error["type"],
                    "command": error["details"]["command"] if "command" in error["details"] else None,
                    "message": error["details"]["message"] if "message" in error["details"] else None,
                    "return_code": error["details"]["return_code"] if "return_code" in error["details"] else None,
                    "stderr": error["details"]["stderr"] if "stderr" in error["details"] else None,
                    "stdout": error["details"]["stdout"] if "stdout" in error["details"] else None,
                    "timeout": error["details"]["timeout"] if "timeout" in error["details"] else None,
                },
            )
        return errors



#!/usr/bin/env python3
# thoth-report-processing
# Copyright(C) 2020 Francesco Murdaca
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Security Indicator aggregator test suite."""

import pytest
import json
from tests.base_test import ReportProcessingTestCase

from thoth.report_processing.components.security import SecurityIndicatorsBandit, SecurityIndicatorsCloc
from thoth.report_processing.components.security import SecurityIndicatorsAggregator

from thoth.report_processing.exceptions import ThothSIPackageNotMatchingException


class TestSecurityReportsBandit(ReportProcessingTestCase):
    """Test implementation of security indicator bandit."""

    _SI_REPORT_NAME = "security-indicator-54c6daf9"

    _SI_FOLDER_PATH = ReportProcessingTestCase.DATA / "security-indicator"

    _SI_AGGREGATOR_REPORTS_FILE = ReportProcessingTestCase.DATA / "results" / "security-aggregated.json"

    def test_create_si_aggregated_dataframe(self) -> None:
        """Test aggregation of SI analuyzers reports for single package."""
        security_aggregator = SecurityIndicatorsAggregator()

        si_bandit_reports = SecurityIndicatorsBandit.aggregate_security_indicator_bandit_results(
            security_indicator_bandit_repo_path=self._SI_FOLDER_PATH,
            is_local=True,
        )
        si_bandit_report = si_bandit_reports[0]

        si_cloc_reports = SecurityIndicatorsCloc.aggregate_security_indicator_cloc_results(
            security_indicator_cloc_repo_path=self._SI_FOLDER_PATH,
            is_local=True,
        )
        si_cloc_report = si_cloc_reports[0]

        si_cloc_report["metadata"]["arguments"]["si-cloc"]["package_name"] = "thoth-test-2"
        with pytest.raises(ThothSIPackageNotMatchingException):
            assert security_aggregator.create_si_aggregated_results(
                si_bandit_report=si_bandit_report,
                si_cloc_report=si_cloc_report,
            )

    def test_create_si_aggregated_json(self) -> None:
        """Test aggregation of SI analuyzers reports for single package."""
        security_aggregator = SecurityIndicatorsAggregator()

        si_bandit_reports = SecurityIndicatorsBandit.aggregate_security_indicator_bandit_results(
            security_indicator_bandit_repo_path=self._SI_FOLDER_PATH,
            is_local=True,
        )
        si_bandit_report = si_bandit_reports[0]

        si_cloc_reports = SecurityIndicatorsCloc.aggregate_security_indicator_cloc_results(
            security_indicator_cloc_repo_path=self._SI_FOLDER_PATH,
            is_local=True,
        )
        si_cloc_report = si_cloc_reports[0]

        aggregated_json = security_aggregator.create_si_aggregated_results(
            si_bandit_report=si_bandit_report,
            si_cloc_report=si_cloc_report,
            output_json=True,
        )

        keys = sorted([k for k in aggregated_json.keys()])

        with open(self._SI_AGGREGATOR_REPORTS_FILE) as json_file:
            aggregated_json_test = json.load(json_file)
        keys_tests = sorted([k for k in aggregated_json_test.keys()])

        assert keys == keys_tests



#!/usr/bin/env python3
# thoth-report-processing
# Copyright(C) 2020 Francesco Murdaca
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Setup for thoth-report-processing."""


import os
from pathlib import Path
from setuptools import find_packages
from setuptools import setup


def _get_install_requires():
    with open("requirements.txt", "r") as requirements_file:
        res = requirements_file.readlines()
        return [req.split(" ", maxsplit=1)[0] for req in res if req]


def _get_version():
    with open(os.path.join("thoth", "report_processing", "__init__.py")) as f:
        content = f.readlines()

    for line in content:
        if line.startswith("__version__ ="):
            # dirty, remove trailing and leading chars
            return line.split(" = ")[1][1:-2]
    raise ValueError("No version identifier found")


VERSION = _get_version()
setup(
    name="thoth-report-processing",
    version=VERSION,
    description="Code for processing report from Thoth components.",
    long_description=Path("README.rst").read_text(),
    long_description_content_type="text/x-rst",
    author="Francesco Murdaca",
    author_email="fmurdaca@redhat.com",
    license="GPLv3+",
    url="https://github.com/thoth-station/report-processing",
    download_url="https://pypi.org/project/thoth-report-processing",
    packages=["thoth.{subpackage}".format(subpackage=p) for p in find_packages("thoth/")],
    include_package_data=True,
    install_requires=_get_install_requires(),
    zip_safe=False,
    command_options={"build_sphinx": {"version": ("setup.py", VERSION), "release": ("setup.py", VERSION)}},
)



#!/usr/bin/env python3
# thoth-report-processing
# Copyright(C) 2020 Francesco Murdaca
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Inspection test suite."""

from tests.base_test import ReportProcessingTestCase

from thoth.report_processing.components.inspection import AmunInspections, AmunInspectionsSummary


class TestAdviser(ReportProcessingTestCase):
    """Test implementation of adviser results."""

    _INSPECTIONS_FOLDER_PATH = ReportProcessingTestCase.DATA / "inspections"

    def test_get_inspection_runs(self) -> None:
        """Test retrieving adviser results from local path."""
        inspection_runs = AmunInspections.aggregate_thoth_inspections_results(
            repo_path=self._INSPECTIONS_FOLDER_PATH,
            is_local=True,
        )
        assert inspection_runs

    def test_create_inspection_summary(self) -> None:
        """Test retrieving adviser results from local path."""
        inspection_runs = AmunInspections.aggregate_thoth_inspections_results(
            repo_path=self._INSPECTIONS_FOLDER_PATH,
            is_local=True,
        )

        processed_inspection_runs, _ = AmunInspections.process_inspection_runs(inspection_runs)
        inspections_df = AmunInspections.create_inspections_dataframe(
            processed_inspection_runs=processed_inspection_runs,
        )

        results, md_report_complete = AmunInspectionsSummary.produce_summary_report(
            inspections_df=inspections_df,
            is_markdown=True,
        )

        assert md_report_complete

    def test_final_dataframe(self) -> None:
        """Test retrieving adviser results from local path."""
        inspection_runs = AmunInspections.aggregate_thoth_inspections_results(
            repo_path=self._INSPECTIONS_FOLDER_PATH,
            is_local=True,
        )
        processed_inspection_runs, _ = AmunInspections.process_inspection_runs(inspection_runs)

        inspections_df = AmunInspections.create_inspections_dataframe(
            processed_inspection_runs=processed_inspection_runs,
        )

        final_dataframe = AmunInspections.create_final_dataframe(inspections_df=inspections_df)
        assert not final_dataframe.empty

    def test_filter_final_dataframe(self) -> None:
        """Test retrieving adviser results from local path."""
        inspection_runs = AmunInspections.aggregate_thoth_inspections_results(
            repo_path=self._INSPECTIONS_FOLDER_PATH,
            is_local=True,
        )
        processed_inspection_runs, _ = AmunInspections.process_inspection_runs(inspection_runs)

        inspections_df = AmunInspections.create_inspections_dataframe(
            processed_inspection_runs=processed_inspection_runs,
        )

        final_dataframe = AmunInspections.create_final_dataframe(inspections_df=inspections_df)

        filtered_df = AmunInspections.filter_final_inspections_dataframe(
            final_inspections_df=final_dataframe,
            pi_name=["PiMatmul"],
            cpus_number=["2"],
            packages={
                "absl-py": ["absl-py-0.9.0-pypi", "absl-py-0.9.0-pypi"],
                "tensorflow-cpu": ["tensorflow-cpu-2.2.0-pypi"],
            },
        )

        assert not filtered_df.empty



#!/usr/bin/env python3
# thoth-report-processing
# Copyright(C) 2020 Francesco Murdaca
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Adviser test suite."""

import pandas as pd
from tests.base_test import ReportProcessingTestCase

from thoth.report_processing.components.adviser import Adviser


class TestAdviser(ReportProcessingTestCase):
    """Test implementation of adviser results."""

    _ADVISER_FOLDER_PATH = ReportProcessingTestCase.DATA / "adviser"

    def test_get_adviser_files(self) -> None:
        """Test retrieving adviser results from local path."""
        adviser_files = Adviser.aggregate_adviser_results(repo_path=self._ADVISER_FOLDER_PATH, is_local=True)
        assert adviser_files

    def test_create_adviser_dataframe_with_results(self) -> None:
        """Test create of adviser dataframe from adviser documents."""
        adviser_files = Adviser.aggregate_adviser_results(repo_path=self._ADVISER_FOLDER_PATH, is_local=True)

        adviser_version = "0.21.1"
        justifications_collected = []
        justifications_collected = Adviser.create_adviser_dataframe(
            adviser_version=adviser_version,
            adviser_files=adviser_files,
            justifications_collected=justifications_collected,
        )

        adviser_dataframe = pd.DataFrame(justifications_collected)

        assert adviser_dataframe.shape[0] == 3

    def test_create_adviser_dataframe_histogram(self) -> None:
        """Test create of adviser dataframe for histogram plot from adviser documents."""
        adviser_files = Adviser.aggregate_adviser_results(repo_path=self._ADVISER_FOLDER_PATH, is_local=True)

        adviser_version = "0.21.1"
        justifications_collected = []
        justifications_collected = Adviser.create_adviser_dataframe(
            adviser_version=adviser_version,
            adviser_files=adviser_files,
            justifications_collected=justifications_collected,
        )

        adviser_dataframe = pd.DataFrame(justifications_collected)

        sorted_justifications_df = Adviser.create_adviser_results_dataframe_histogram(
            adviser_type_dataframe=adviser_dataframe,
        )

        assert sorted_justifications_df.shape[0] == 3

    def test_create_adviser_dataframe_heatmap(self) -> None:
        """Test create of adviser dataframe for heatmap plot from adviser documents."""
        adviser_files = Adviser.aggregate_adviser_results(repo_path=self._ADVISER_FOLDER_PATH, is_local=True)

        adviser_version = "0.21.1"
        justifications_collected = []
        justifications_collected = Adviser.create_adviser_dataframe(
            adviser_version=adviser_version,
            adviser_files=adviser_files,
            justifications_collected=justifications_collected,
        )

        adviser_dataframe = pd.DataFrame(justifications_collected)

        adviser_heatmap_df = Adviser.create_adviser_results_dataframe_heatmap(
            adviser_type_dataframe=adviser_dataframe,
            number_days=1,
        )

        last_date = [column for column in adviser_heatmap_df.columns][-1]
        csv = adviser_heatmap_df[[last_date]].to_csv(header=False)

        assert csv



#!/usr/bin/env python3
# thoth-report-processing
# Copyright(C) 2020 Francesco Murdaca
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Thoth components classes to process outputs."""

from .adviser import Adviser
from .inspection import AmunInspections
from .inspection import AmunInspectionsSummary
from .inspection import AmunInspectionsStatistics
from .security import SecurityIndicatorsBandit
from .security import SecurityIndicatorsCloc
from .security import SecurityIndicatorsAggregator
from .solver import Solver

__all__ = [
    "Adviser",
    "AmunInspections",
    "AmunInspectionsStatistics",
    "AmunInspectionsSummary",
    "SecurityIndicatorsBandit",
    "SecurityIndicatorsCloc",
    "SecurityIndicatorsAggregator",
    "Solver",
]



#!/usr/bin/env python3
# thoth-report-processing
# Copyright(C) 2020 Francesco Murdaca
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Solver test suite."""

from tests.base_test import ReportProcessingTestCase

from thoth.report_processing.components.solver import Solver


class TestSolver(ReportProcessingTestCase):
    """Test implementation of solver results."""

    _SOLVER_FOLDER_PATH = ReportProcessingTestCase.DATA / "solver"

    def test_get_solver_reports(self) -> None:
        """Test retrieving solver results from local path."""
        solver_reports = Solver.aggregate_solver_results(repo_path=self._SOLVER_FOLDER_PATH, is_local=True)
        assert solver_reports

    def test_get_metadata_df_from_solver_report(self) -> None:
        """Test obtaining metadata from si solver report."""
        solver_reports = Solver.aggregate_solver_results(repo_path=self._SOLVER_FOLDER_PATH, is_local=True)
        solver_report = solver_reports["solver-fedora-31-py37-012b745d"]
        metadata_retrieved = Solver.extract_data_from_solver_metadata(solver_report_metadata=solver_report["metadata"])
        metadata_retrieved_keys = [k for k in metadata_retrieved]
        metadata_test_keys = [
            "document_id",
            "datetime",
            "requirements",
            "solver",
            "os_name",
            "os_version",
            "python_interpreter",
            "analyzer_version",
        ]
        assert metadata_retrieved_keys == metadata_test_keys
